"""
resolvers.py
---------------

Provides a common interface to load assets referenced by name
like MTL files, texture images, etc. Assets can be from ZIP
archives, web assets, or a local file path.
"""

import os
import abc
import itertools

from . import util
from . import caching

# URL parsing for remote resources via WebResolver
try:
    # Python 3
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    from urlparse import urlparse


class Resolver(util.ABC):
    """
    The base class for resolvers.
    """

    @abc.abstractmethod
    def __init__(self, *args, **kwargs):
        raise NotImplementedError('Use a resolver subclass!')

    @abc.abstractmethod
    def get(self, key):
        raise NotImplementedError()

    @abc.abstractmethod
    def write(self, name, data):
        raise NotImplementedError('`write` not implemented!')

    @abc.abstractmethod
    def namespaced(self, namespace):
        raise NotImplementedError('`namespaced` not implemented!')

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.write(key, value)

    def __contains__(self, key):
        return key in self.keys()


class FilePathResolver(Resolver):
    """
    Resolve files from a source path on the file system.
    """

    def __init__(self, source):
        """
        Resolve files based on a source path.

        Parameters
        ------------
        source : str
          File path where mesh was loaded from
        """
        # remove everything other than absolute path
        clean = os.path.expanduser(
            os.path.abspath(str(source)))

        self.clean = clean
        if os.path.isdir(clean):
            # if we were passed a directory use it
            self.parent = clean
        else:
            # otherwise get the parent directory we've been passed
            split = os.path.split(clean)
            self.parent = split[0]

        # exit if directory doesn't exist
        if not os.path.isdir(self.parent):
            raise ValueError(
                'path `{} `not a directory!'.format(
                    self.parent))

    def keys(self):
        """
        List all files available to be loaded.

        Yields
        -----------
        name : str
          Name of a file which can be accessed.
        """
        parent = self.parent
        for path, _, names in os.walk(self.parent):
            # strip any leading parent key
            if path.startswith(parent):
                path = path[len(parent):]
            # yield each name
            for name in names:
                yield os.path.join(path, name)

    def namespaced(self, namespace):
        """
        Return a resolver which changes the root of the
        resolver by an added namespace.

        Parameters
        -------------
        namespace : str
          Probably a subdirectory

        Returns
        --------------
        resolver : FilePathResolver
          Resolver with root directory changed.
        """
        return FilePathResolver(os.path.join(
            self.parent, namespace))

    def get(self, name):
        """
        Get an asset.

        Parameters
        -------------
        name : str
          Name of the asset

        Returns
        ------------
        data : bytes
          Loaded data from asset
        """
        # load the file by path name
        path = os.path.join(self.parent, name.strip())
        if not os.path.exists(path):
            path = os.path.join(
                self.parent, os.path.split(name)[-1])
        with open(path, 'rb') as f:
            data = f.read()
        return data

    def write(self, name, data):
        """
        Write an asset to a file path.

        Parameters
        -----------
        name : str
          Name of the file to write
        data : str or bytes
          Data to write to the file
        """
        # write files to path name
        with open(os.path.join(self.parent, name.strip()), 'wb') as f:
            # handle encodings correctly for str/bytes
            util.write_encoded(file_obj=f, stuff=data)


class ZipResolver(Resolver):
    """
    Resolve files inside a ZIP archive.
    """

    def __init__(self, archive=None, namespace=None):
        """
        Resolve files inside a ZIP archive as loaded by
        trimesh.util.decompress

        Parameters
        -------------
        archive : dict
          Contains resources as file object
        namespace : None or str
          If passed will only show keys that start
          with this value and this substring must be
          removed for any get calls.
        """
        self.archive = archive
        if isinstance(namespace, str):
            self.namespace = namespace.strip().rstrip('/') + '/'
        else:
            self.namespace = None

    def keys(self):
        """
        Get the available keys in the current archive.

        Returns
        -----------
        keys : iterable
          Keys in the current archive.
        """
        if self.namespace is not None:
            namespace = self.namespace
            length = len(namespace)
            # only return keys that start with the namespace
            # and strip off the namespace from the returned
            # keys.
            return [k[length:] for k in self.archive.keys()
                    if k.startswith(namespace)
                    and len(k) > length]
        return self.archive.keys()

    def write(self, key, value):
        """
        Store a value in the current archive.

        Parameters
        -----------
        key : hashable
          Key to store data under.
        value : str, bytes, file-like
          Value to store.
        """
        if self.archive is None:
            self.archive = {}
        self.archive[key] = value

    def get(self, name):
        """
        Get an asset from the ZIP archive.

        Parameters
        -------------
        name : str
          Name of the asset

        Returns
        -------------
        data : bytes
          Loaded data from asset
        """
        # not much we can do with None
        if name is None:
            return
        # make sure name is a string
        if hasattr(name, 'decode'):
            name = name.decode('utf-8')
        # store reference to archive inside this function
        archive = self.archive
        # requested name not identical in
        # storage so attempt to recover
        if name not in archive:
            # loop through unique results
            for option in nearby_names(name, self.namespace):
                if option in archive:
                    # cleaned option is in archive
                    # so store value and exit
                    name = option
                    break

        # get the stored data
        obj = archive[name]
        # if the dict is storing data as bytes just return
        if isinstance(obj, (bytes, str)):
            return obj
        # otherwise get it as a file object
        # read file object from beginning
        obj.seek(0)
        # data is stored as a file object
        data = obj.read()
        obj.seek(0)
        return data

    def namespaced(self, namespace):
        """
        Return a "sub-resolver" with a root namespace.

        Parameters
        -------------
        namespace : str
          The root of the key to clip off, i.e. if
          this resolver has key `a/b/c` you can get
          'a/b/c' with resolver.namespaced('a/b').get('c')

        Returns
        -----------
        resolver : Resolver
          Namespaced resolver.
        """
        return ZipResolver(archive=self.archive,
                           namespace=namespace)

    def export(self):
        """
        Export the contents of the current archive as
        a ZIP file.

        Returns
        ------------
        compressed : bytes
          Compressed data in ZIP format.
        """
        return util.compress(self.archive)


class WebResolver(Resolver):
    """
    Resolve assets from a remote URL.
    """

    def __init__(self, url):
        """
        Resolve assets from a base URL.

        Parameters
        --------------
        url : str
          Location where a mesh was stored or
          directory where mesh was stored
        """
        if hasattr(url, 'decode'):
            url = url.decode('utf-8')

        # parse string into namedtuple
        parsed = urlparse(url)

        # we want a base url
        split = [i for i in parsed.path.split('/')
                 if len(i) > 0]

        # if the last item in the url path is a filename
        # move up a "directory" for the base path
        if len(split) == 0:
            path = ''
        elif '.' in split[-1]:
            # clip off last item
            path = '/'.join(split[:-1])
        else:
            # recombine into string ignoring any double slashes
            path = '/'.join(split)
        self.base_url = '/'.join(i for i in [
            parsed.scheme + ':/',
            parsed.netloc.strip('/'),
            path.strip('/')] if len(i) > 0) + '/'

        # our string handling should have never inserted double slashes
        assert '//' not in self.base_url[len(parsed.scheme) + 3:]
        # we should always have ended with a single slash
        assert self.base_url.endswith('/')

    def get(self, name):
        """
        Get a resource from the remote site.

        Parameters
        -------------
        name : str
          Asset name, i.e. 'quadknot.obj.mtl'
        """
        # do import here to keep soft dependency
        import requests

        # remove leading and trailing whitespace
        name = name.strip()
        # fetch the data from the remote url

        # base url has been carefully formatted
        url = self.base_url + name

        response = requests.get(url)

        if response.status_code != 200:
            # try to strip off filesystem crap
            if name.startswith('./'):
                name = name[2:]
            response = requests.get(self.base_url + name)

        if response.status_code == '404':
            raise ValueError(response.content)

        # return the bytes of the response
        return response.content

    def namespaced(self, namespace):
        """
        Return a namespaced version of current resolver.

        Parameters
        -------------
        namespace : str
          URL fragment

        Returns
        -----------
        resolver : WebResolver
          With sub-url: `https://example.com/{namespace}`
        """
        # join the base url and the namespace
        return WebResolver(url=self.base_url + namespace)

    def write(self, key, value):
        raise NotImplementedError("can't write to remote")


class GithubResolver(Resolver):
    def __init__(self,
                 repo,
                 branch=None,
                 commit=None,
                 save=None):
        """
        Get files from a remote Github repository by
        downloading a zip file with the entire branch
        or a specific commit.

        Parameters
        -------------
        repo : str
          In the format of `owner/repo`
        branch : str
          The remote branch you want to get files from.
        commit : str
          The full commit hash: pass either this OR branch.
        save : None or str
          A path if you want to save results locally.
        """
        # the github URL for the latest commit of a branch.
        if commit is None:
            self.url = (
                'https://github.com/{repo}/archive/' +
                'refs/heads/{branch}.zip').format(
                    repo=repo, branch=branch)
        else:
            # get a commit URL
            self.url = ('https://github.com/{repo}/archive/' +
                        '{commit}.zip').format(
                            repo=repo, commit=commit)

        if save is not None:
            self.cache = caching.DiskCache(save)
        else:
            self.cache = None

    def keys(self):
        """
        List the available files in the repository.

        Returns
        ----------
        keys : iterable
          Keys available to the resolved.
        """
        return self.zipped.keys()

    def write(self, name, data):
        raise NotImplementedError('`write` not implemented!')

    @property
    def zipped(self):
        """

        - opened zip file
        - locally saved zip file
        - retrieve zip file and saved
        """
        def fetch():
            """
            Fetch the remote zip file.
            """
            import requests
            response = requests.get(self.url)
            if not response.ok:
                raise ValueError(response.content)
            return response.content

        if hasattr(self, '_zip'):
            return self._zip
        # download the archive or get from disc
        raw = self.cache.get(self.url, fetch)
        # create a zip resolver for the archive
        self._zip = ZipResolver(util.decompress(
            util.wrap_as_stream(raw), file_type='zip'))

        return self._zip

    def get(self, key):
        return self.zipped.get(key)

    def namespaced(self, namespace):
        """
        Return a "sub-resolver" with a root namespace.

        Parameters
        -------------
        namespace : str
          The root of the key to clip off, i.e. if
          this resolver has key `a/b/c` you can get
          'a/b/c' with resolver.namespaced('a/b').get('c')

        Returns
        -----------
        resolver : Resolver
          Namespaced resolver.
        """
        return self.zipped.namespaced(namespace)


def nearby_names(name, namespace=None):
    """
    Try to find nearby variants of a specified name.

    Parameters
    ------------
    name : str
      Initial name.

    Yields
    -----------
    nearby : str
      Name that is a lightly permutated version
      of the initial name.
    """
    # the various operations that *might* result in a correct key
    def trim(prefix, item):
        if item.startswith(prefix):
            return item[len(prefix):]
        return item

    cleaners = [lambda x: x,
                lambda x: x.strip(),
                lambda x: trim('./', x),
                lambda x: trim('.\\', x),
                lambda x: trim('\\', x),
                lambda x: os.path.split(x)[-1],
                lambda x: x.replace('%20', ' ')]

    if namespace is None:
        namespace = ''

    # make sure we don't return repeat values
    hit = set()
    for f in cleaners:
        # try just one cleaning function
        current = f(name)
        if current in hit:
            continue
        hit.add(current)
        yield namespace + current

    for a, b in itertools.combinations(cleaners, 2):
        # apply both clean functions
        current = a(b(name))
        if current in hit:
            continue
        hit.add(current)
        yield namespace + current

        # try applying in reverse order
        current = b(a(name))
        if current in hit:
            continue
        hit.add(current)
        yield namespace + current

    if '..' in name and namespace is not None:
        # if someone specified relative paths give it one attempt
        strip = namespace.strip('/').split('/')[:-name.count('..')]
        strip.extend(name.split('..')[-1].strip('/').split('/'))
        yield '/'.join(strip)



try:
    from . import generic as g
except BaseException:
    import generic as g


class NSphereTest(g.unittest.TestCase):

    def test_minball(self):
        # how close do we need to be
        tol_fit = 1e-2

        # get some assorted mesh geometries to test performance
        # and a perfect sphere mesh to test the degenerate case
        for m in g.np.append(list(g.get_meshes(5)),
                             g.trimesh.primitives.Sphere()):

            s = m.bounding_sphere
            R_check = ((m.vertices - s.primitive.center)
                       ** 2).sum(axis=1).max() ** .5

            assert len(s.primitive.center) == 3
            assert s.primitive.radius > 0.0
            assert abs(s.primitive.radius - R_check) < tol_fit
            assert s.volume > (m.volume - tol_fit)

        # check minimum n-sphere for points in 2, 3, 4 dimensions
        for d in [2, 3, 4]:
            for _i in range(5):
                points = g.random((100, d))
                C, R = g.trimesh.nsphere.minimum_nsphere(points)
                R_check = ((points - C)**2).sum(axis=1).max() ** .5
                assert len(C) == d
                assert R > 0.0
                assert abs(R - R_check) < g.tol.merge

    def test_isnsphere(self):
        # make sure created spheres are uv sphere
        m = g.trimesh.creation.uv_sphere()
        # move the mesh around for funsies
        m.apply_translation(g.random(3))
        m.apply_transform(
            g.trimesh.transformations.random_rotation_matrix())
        # all vertices should be on nsphere
        assert g.trimesh.nsphere.is_nsphere(m.vertices)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class SplineTests(g.unittest.TestCase):

    def test_bezier_example(self):
        # path with a bunch of bezier spline
        p = g.get_mesh('2D/MIL.svg')
        # should have one body
        assert len(p.polygons_full) == 1

        # length of perimeter of exterior
        truth = 12696.6

        # perimeter should be about right if it was discretized properly
        if not g.np.isclose(p.polygons_full[0].exterior.length,
                            truth,
                            atol=100.0):
            raise ValueError('perimeter wrong: {} != {}'.format(
                truth, p.polygons_full[0].exterior.length))


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
test_upstream.py
-------------------

Make sure our upstream dependencies are acting right.
"""

try:
    from . import generic as g
except BaseException:
    import generic as g


class UpstreamTests(g.unittest.TestCase):
    def test_shapely(self):
        """
        conda installs of shapely started returning NaN on
        valid input so make sure our builds fail in that case
        """
        string = g.LineString([[0, 0], [1, 0]])
        assert g.np.isclose(string.length, 1.0)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class MFTest(g.unittest.TestCase):

    def test_3MF(self):
        # an assembly with instancing
        s = g.get_mesh('counterXP.3MF')

        # should be 2 unique meshes
        assert len(s.geometry) == 2
        # should be 6 instances around the scene
        assert len(s.graph.nodes_geometry) == 6
        assert all(m.is_volume for m in s.geometry.values())

        # a single body 3MF assembly
        s = g.get_mesh('featuretype.3MF')
        # should be 2 unique meshes
        assert len(s.geometry) == 1
        # should be 6 instances around the scene
        assert len(s.graph.nodes_geometry) == 1

    def test_units(self):
        # test our unit conversion function
        converter = g.trimesh.units.unit_conversion
        # these are the units listed in the 3MF spec as valid
        units = ['micron', 'millimeter',
                 'centimeter', 'inch', 'foot', 'meter']
        # check conversion factor for all valid 3MF units
        assert all(converter(u, 'inches') > 1e-12 for u in units)

    def test_kwargs(self):
        # check if kwargs are properly passed to geometries
        s = g.get_mesh('P_XPM_0331_01.3mf')
        assert all(len(v.vertices) == 4 for v in s.geometry.values())

        s = g.get_mesh('P_XPM_0331_01.3mf', process=False)
        assert all(len(v.vertices) == 5 for v in s.geometry.values())

    def test_names(self):
        # check if two different objects with the same name are correctly
        # processed
        s = g.get_mesh('cube_and_sphere_same_name.3mf')
        assert len(s.geometry) == 2

    def test_roundtrip(self):
        if g.sys.version_info < (3, 6):
            g.log.warning('relies on > Python 3.5')
            return

        # test a scene round-tripped through the
        # 3MF exporter and importer
        s = g.get_mesh('cycloidal.3DXML')
        assert len(s.geometry) == 13

        # export and reload
        r = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(
                s.export(file_type='3mf')),
            file_type='3mf')

        assert set(s.geometry.keys()) == set(r.geometry.keys())
        assert g.np.allclose(s.bounds, r.bounds)
        assert g.np.isclose(s.area, r.area, rtol=1e-3)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class MargeTest(g.unittest.TestCase):

    def test_cube(self):
        """
        Test PointCloud object
        """

        m = g.trimesh.creation.box()

        assert m.vertices.shape == (8, 3)
        assert m.is_volume
        assert m.euler_number == 2

        # stack a bunch of unreferenced vertices
        m.vertices = g.np.vstack((
            m.vertices, g.random((10000, 3))))
        assert m.euler_number == 2
        assert m.vertices.shape == (10008, 3)
        assert m.referenced_vertices.sum() == 8
        copied = m.copy()

        # should remove unreferenced vertices
        m.merge_vertices()
        assert len(m.vertices) == 8
        assert m.is_volume
        assert m.euler_number == 2
        assert m.referenced_vertices.sum() == 8

        m.update_vertices(m.referenced_vertices)
        assert len(m.vertices) == 8
        assert m.is_volume
        assert m.euler_number == 2
        assert m.referenced_vertices.sum() == 8

        # check the copy with an int mask and see if inverse is created
        assert copied.referenced_vertices.sum() == 8
        assert copied.vertices.shape == (10008, 3)
        mask = g.np.nonzero(copied.referenced_vertices)[0]

        copied.update_vertices(mask)
        assert len(copied.vertices) == 8
        assert copied.is_volume
        assert copied.euler_number == 2
        assert copied.referenced_vertices.sum() == 8


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class RepairTests(g.unittest.TestCase):

    def test_fill_holes(self):
        for mesh_name in ['unit_cube.STL',
                          'machinist.XAML',
                          'round.stl',
                          'sphere.ply',
                          'teapot.stl',
                          'soup.stl',
                          'featuretype.STL',
                          'angle_block.STL',
                          'quadknot.obj']:

            mesh = g.get_mesh(mesh_name)
            if not mesh.is_watertight:
                # output of fill_holes should match watertight status
                returned = mesh.fill_holes()
                assert returned == mesh.is_watertight
                continue

            hashes = [{mesh._data.__hash__(),
                       mesh._data.__hash__(),
                       mesh._data.fast_hash()}]

            mesh.faces = mesh.faces[1:-1]
            assert not mesh.is_watertight
            assert not mesh.is_volume

            # color some faces
            g.trimesh.repair.broken_faces(
                mesh, color=[255, 0, 0, 255])

            hashes.append({mesh._data.__hash__(),
                           mesh._data.__hash__(),
                           mesh._data.fast_hash()})

            assert hashes[0] != hashes[1]

            # run the fill holes operation should succeed
            assert mesh.fill_holes()
            # should be a superset of the last two
            assert mesh.is_volume
            assert mesh.is_watertight
            assert mesh.is_winding_consistent

            hashes.append({mesh._data.__hash__(),
                           mesh._data.__hash__(),
                           mesh._data.fast_hash()})
            assert hashes[1] != hashes[2]

            # try broken faces on a watertight mesh
            g.trimesh.repair.broken_faces(
                mesh, color=[255, 255, 0, 255])

    def test_fix_normals(self):
        for mesh in g.get_meshes(5):
            mesh.fix_normals()

    def test_winding(self):
        """
        Reverse some faces and make sure fix_face_winding flips
        them back.
        """

        meshes = [g.get_mesh(i) for i in
                  ['unit_cube.STL',
                   'machinist.XAML',
                   'round.stl',
                   'quadknot.obj',
                   'soup.stl']]

        for i, mesh in enumerate(meshes):
            # turn scenes into multibody meshes
            if g.trimesh.util.is_instance_named(mesh, 'Scene'):
                meta = mesh.metadata
                meshes[i] = mesh.dump().sum()
                meshes[i].metadata = meta

        timing = {}
        for mesh in meshes:
            # save the initial state
            is_volume = mesh.is_volume
            winding = mesh.is_winding_consistent

            tic = g.time.time()
            # flip faces to break winding
            mesh.faces[:4] = g.np.fliplr(mesh.faces[:4])

            # run the operation
            mesh.fix_normals()

            # make sure mesh is repaired to former glory
            assert mesh.is_volume == is_volume
            assert mesh.is_winding_consistent == winding

            # save timings
            timing[mesh.metadata['file_name']] = g.time.time() - tic
        # print timings as a warning
        g.log.warning(g.json.dumps(timing, indent=4))

    def test_inversion(self):
        """Make sure fix_inversion switches all reversed faces back"""
        orig_mesh = g.get_mesh("unit_cube.STL")
        orig_verts = orig_mesh.vertices.copy()
        orig_faces = orig_mesh.faces.copy()

        mesh = g.Trimesh(orig_verts, orig_faces[:, ::-1])
        inv_faces = mesh.faces.copy()
        # check not fixed on the way in
        assert not g.np.allclose(inv_faces, orig_faces)

        g.trimesh.repair.fix_inversion(mesh)
        assert not g.np.allclose(mesh.faces, inv_faces)
        assert g.np.allclose(mesh.faces, orig_faces)

    def test_multi(self):
        """
        Try repairing a multibody geometry
        """
        # create a multibody mesh with two cubes
        a = g.get_mesh('unit_cube.STL')
        b = a.copy()
        b.apply_translation([2, 0, 0])
        m = a + b
        # should be a volume: watertight, correct winding
        assert m.is_volume

        # flip one face of A
        a.faces[:1] = g.np.fliplr(a.faces[:1])
        # flip every face of A
        a.invert()
        # flip one face of B
        b.faces[:1] = g.np.fliplr(b.faces[:1])
        m = a + b

        # not a volume
        assert not m.is_volume

        m.fix_normals(multibody=False)

        # shouldn't fix inversion of one cube
        assert not m.is_volume

        # run fix normal with multibody mode
        m.fix_normals()

        # should be volume again
        assert m.is_volume

        # mesh should be volume of two boxes, and positive
        assert g.np.isclose(m.volume, 2.0)

    def test_flip(self):
        # create two spheres
        a = g.trimesh.creation.icosphere()
        b = g.trimesh.creation.icosphere().apply_translation([2, 3, 0])
        # invert the second sphere
        b.faces = g.np.fliplr(b.faces)
        m = a + b
        # make sure normals are in cache
        assert m.face_normals.shape == m.faces.shape
        m.fix_normals(multibody=True)
        assert g.np.isclose(m.volume, a.volume * 2.0)

    def test_fan(self):

        # start by creating an icosphere and removing
        # all faces that include a single vertex to make
        # a nice hole in the mesh
        m = g.trimesh.creation.icosphere()
        clip = m.vertex_faces[0]
        clip = clip[clip >= 0]
        assert len(clip) > 4
        mask = g.np.ones(len(m.faces), dtype=bool)
        mask[clip] = False

        # should have been watertight
        assert m.is_watertight
        assert m.is_winding_consistent
        m.update_faces(mask)
        # now should not be watertight
        assert not m.is_watertight
        assert m.is_winding_consistent

        # create a triangle fan to cover the hole
        stitch = g.trimesh.repair.stitch(m)
        # should be an (n, 3) int
        assert len(stitch.shape) == 2
        assert stitch.shape[1] == 3
        assert stitch.dtype.kind == 'i'

        # now check our stitch to see if it handled the hole
        repair = g.trimesh.Trimesh(
            vertices=m.vertices.copy(),
            faces=g.np.vstack((m.faces, stitch)))
        assert repair.is_watertight
        assert repair.is_winding_consistent


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import os
import json

from ..util import decode_text, wrap_as_stream

# find the current absolute path to this directory
_pwd = os.path.expanduser(os.path.abspath(
    os.path.dirname(__file__)))

# once resources are loaded cache them
_cache = {}


def get(name, decode=True, decode_json=False, as_stream=False):
    """
    Get a resource from the `trimesh/resources` folder.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`
    decode : bool
      Whether or not to decode result as UTF-8
    decode_json : bool
      Run `json.loads` on resource if True.
    as_stream : bool
      Return as a file-like object

    Returns
    -------------
    resource : str, bytes, or decoded JSON
      File data
    """
    # key by name and decode
    cache_key = (name,
                 bool(decode),
                 bool(decode_json),
                 bool(as_stream))
    cached = _cache.get(cache_key)
    if hasattr(cached, 'seek'):
        cached.seek(0)
    if cached is not None:
        return cached

    # get the resource using relative names
    with open(os.path.join(_pwd, name), 'rb') as f:
        resource = f.read()

    # make sure we return it as a string if asked
    if decode:
        # will decode into text if possibly
        resource = decode_text(resource)

    if decode_json:
        resource = json.loads(resource)
    elif as_stream:
        resource = wrap_as_stream(resource)

    # store for later access
    _cache[cache_key] = resource

    return resource


def get_schema(name):
    """
    Load a schema and evaluate the referenced files.

    Parameters
    ------------
    name : str
      Filename of schema.

    Returns
    ----------
    schema : dict
      Loaded and resolved schema.
    """
    from ..schemas import resolve
    from ..resolvers import FilePathResolver
    # get a resolver for our base path
    resolver = FilePathResolver(
        os.path.join(_pwd, 'schema', name))
    # recursively load $ref keys
    schema = resolve(
        json.loads(decode_text(resolver.get(name))),
        resolver=resolver)
    return schema



import trimesh
import unittest
import logging
import numpy as np

try:
    from . import generic as g
except BaseException:
    import generic as g

TEST_DIM = (100, 3)
TOL_ZERO = 1e-9
TOL_CHECK = 1e-2

log = logging.getLogger('trimesh')
log.addHandler(logging.NullHandler())


class VectorTests(unittest.TestCase):

    def setUp(self):
        self.test_dim = TEST_DIM

    def test_unitize_multi(self):
        vectors = np.ones(self.test_dim)
        vectors[0] = [0, 0, 0]
        vectors, valid = trimesh.unitize(vectors, check_valid=True)

        assert not valid[0]
        assert valid[1:].all()

        length = np.sum(vectors[1:] ** 2, axis=1) ** .5
        assert np.allclose(length, 1.0)

    def test_align(self):
        log.info('Testing vector alignment')
        target = np.array([0, 0, 1])
        for _i in range(100):
            vector = trimesh.unitize(np.random.random(3) - .5)
            T = trimesh.geometry.align_vectors(vector, target)
            result = np.dot(T, np.append(vector, 1))[0:3]
            aligned = np.abs(result - target).sum() < TOL_ZERO
            self.assertTrue(aligned)


class UtilTests(unittest.TestCase):

    def test_bounds_tree(self):
        for _attempt in range(3):
            for dimension in [2, 3]:
                t = g.random((1000, 3, dimension))
                bounds = g.np.column_stack((t.min(axis=1), t.max(axis=1)))
                tree = g.trimesh.util.bounds_tree(bounds)
                self.assertTrue(0 in tree.intersection(bounds[0]))

    def test_stack(self):
        # shortcut to the function
        f = g.trimesh.util.stack_3D
        # start with some random points
        p = g.random((100, 2))
        stack = f(p)
        # shape should be 3D
        assert stack.shape == (100, 3)
        # points should be equal
        assert g.np.allclose(p, stack[:, :2])

        # check an empty array
        assert f([]).shape == (0,)

        try:
            # try with 4D points
            f(g.np.ones((100, 4)))
            raise AssertionError()
        except ValueError:
            # this is what should happen
            pass

    def test_has_module(self):

        assert g.trimesh.util.has_module('collections')
        assert not g.trimesh.util.has_module('foobarrionananan')

    def test_strips(self):
        """
        Test our conversion of triangle strips to face indexes.
        """

        def strips_to_faces(strips):
            """
            A slow but straightforward version of the function to test against
            """
            faces = g.collections.deque()
            for s in strips:
                s = g.np.asanyarray(s, dtype=g.np.int64)
                # each triangle is defined by one new vertex
                tri = g.np.column_stack([g.np.roll(s, -i)
                                         for i in range(3)])[:-2]
                # we need to flip ever other triangle
                idx = (g.np.arange(len(tri)) % 2).astype(bool)
                tri[idx] = g.np.fliplr(tri[idx])
                faces.append(tri)
            # stack into one (m,3) array
            faces = g.np.vstack(faces)
            return faces

        # test 4- triangle strip
        s = [g.np.arange(6)]
        f = g.trimesh.util.triangle_strips_to_faces(s)
        assert (f == g.np.array([[0, 1, 2],
                                 [3, 2, 1],
                                 [2, 3, 4],
                                 [5, 4, 3]])).all()
        assert len(f) + 2 == len(s[0])
        assert (f == strips_to_faces(s)).all()

        # test single triangle
        s = [g.np.arange(3)]
        f = g.trimesh.util.triangle_strips_to_faces(s)
        assert (f == g.np.array([[0, 1, 2]])).all()
        assert len(f) + 2 == len(s[0])
        assert (f == strips_to_faces(s)).all()

        s = [g.np.arange(100)]
        f = g.trimesh.util.triangle_strips_to_faces(s)
        assert len(f) + 2 == len(s[0])
        assert (f == strips_to_faces(s)).all()

    def test_pairwise(self):
        # check to make sure both itertools and numpy
        # methods return the same result
        pa = np.array(list(g.trimesh.util.pairwise(range(5))))
        pb = g.trimesh.util.pairwise(np.arange(5))

        # make sure results are the same from both methods
        assert (pa == pb).all()
        # make sure we have 4 pairs for 5 values
        assert len(pa) == 4
        # make sure all pairs are length 2
        assert all(len(i) == 2 for i in pa)

    def test_concat(self):

        a = g.get_mesh('ballA.off')
        b = g.get_mesh('ballB.off')

        hA = a.__hash__()
        hB = b.__hash__()

        # make sure we're not mutating original mesh
        for _i in range(4):
            c = a + b
            assert g.np.isclose(c.volume,
                                a.volume + b.volume)
            assert a.__hash__() == hA
            assert b.__hash__() == hB

        count = 5
        meshes = []
        for _i in range(count):
            m = a.copy()
            m.apply_translation([a.scale, 0, 0])
            meshes.append(m)

        # do a multimesh concatenate
        r = g.trimesh.util.concatenate(meshes)
        assert g.np.isclose(r.volume,
                            a.volume * count)
        assert a.__hash__() == hA

    def test_unique_id(self):
        num_ids = 10000

        g.trimesh.util.random.seed(0)
        unique_ids_0 = []
        for _i in range(num_ids):
            s = g.trimesh.util.unique_id()
            unique_ids_0.append(s)

        # make sure every id is truly unique
        assert len(unique_ids_0) == len(g.np.unique(unique_ids_0))

        g.trimesh.util.random.seed(0)
        unique_ids_1 = []
        for i in range(num_ids):
            s = g.trimesh.util.unique_id()
            unique_ids_1.append(s)

            # make sure id's can be reproduced
            assert s == unique_ids_0[i]

    def test_unique_name(self):
        from trimesh.util import unique_name

        assert len(unique_name(None, {})) > 0
        assert len(unique_name('', {})) > 0

        count = 10
        names = set()
        for _i in range(count):
            names.add(unique_name('hi', names))
        assert len(names) == count

        names = set()
        for _i in range(count):
            names.add(unique_name('', names))
        assert len(names) == count

        # Try with a larger set of names
        # get some random strings
        names = [g.uuid4().hex for _ in range(20)]
        # make it a whole lotta duplicates
        names = names * 1000
        # add a non-int postfix to test
        names.extend(['suppp_hi'] * 10)

        assigned = set()
        with g.Profiler() as P:
            for name in names:
                assigned.add(unique_name(name, assigned))
        g.log.debug(P.output_text())

        assigned_new = set()
        # tracker = UniqueName()\
        counts = {}
        with g.Profiler() as P:
            for name in names:
                assigned_new.add(unique_name(
                    name,
                    contains=assigned_new,
                    counts=counts))
        g.log.debug(P.output_text())

        # new scheme should match the old one
        assert assigned_new == assigned
        # de-duplicated set should match original length
        assert len(assigned) == len(names)


class ContainsTest(unittest.TestCase):

    def test_inside(self):
        sphere = g.trimesh.primitives.Sphere(radius=1.0, subdivisions=4)
        g.log.info('Testing contains function with sphere')
        samples = (np.random.random((1000, 3)) - .5) * 5
        radius = np.linalg.norm(samples, axis=1)

        margin = .05
        truth_in = radius < (1.0 - margin)
        truth_out = radius > (1.0 + margin)

        contains = sphere.contains(samples)

        if not contains[truth_in].all():
            raise ValueError('contains test does not match truth!')

        if contains[truth_out].any():
            raise ValueError('contains test does not match truth!')


class IOWrapTests(unittest.TestCase):

    def test_io_wrap(self):

        util = g.trimesh.util

        # check wrap_as_stream
        test_b = g.random(1).tobytes()
        test_s = 'this is a test yo'
        res_b = util.wrap_as_stream(test_b).read()
        res_s = util.wrap_as_stream(test_s).read()
        assert res_b == test_b
        assert res_s == test_s

        # check __enter__ and __exit__
        hi = 'hi'.encode('utf-8')
        with util.BytesIO(hi) as f:
            assert f.read() == hi

        # check __enter__ and __exit__
        hi = 'hi'
        with util.StringIO(hi) as f:
            assert f.read() == hi


class CompressTests(unittest.TestCase):

    def test_compress(self):

        source = {'hey': 'sup',
                  'naa': '2002211'}

        # will return bytes
        c = g.trimesh.util.compress(source)

        # wrap bytes as file- like object
        f = g.trimesh.util.wrap_as_stream(c)
        # try to decompress file- like object
        d = g.trimesh.util.decompress(f, file_type='zip')

        # make sure compressed- decompressed items
        # are the same after a cycle
        for key, value in source.items():
            result = d[key].read().decode('utf-8')
            assert result == value


class UniqueTests(unittest.TestCase):

    def test_unique(self):

        options = [np.array([0, 1, 2, 3, 1, 3, 10, 20]),
                   np.arange(100),
                   np.array([], dtype=np.int64),
                   (np.random.random(1000) * 10).astype(int)]

        for values in options:
            if len(values) > 0:
                minlength = values.max()
            else:
                minlength = 10

            # try our unique bincount function
            unique, inverse, counts = g.trimesh.grouping.unique_bincount(
                values,
                minlength=minlength,
                return_inverse=True,
                return_counts=True)
            # make sure inverse is correct
            assert (unique[inverse] == values).all()

            # make sure that the number of counts matches
            # the number of unique values
            assert (len(unique) == len(counts))

            # get the truth
            truth_unique, truth_inverse, truth_counts = np.unique(values,
                                                                  return_inverse=True,
                                                                  return_counts=True)
            # make sure truth is doing what we think
            assert (truth_unique[truth_inverse] == values).all()

            # make sure we have same number of values
            assert len(truth_unique) == len(unique)

            # make sure all values are identical
            assert set(truth_unique) == set(unique)

            # make sure that the truth counts are identical to our counts
            assert np.all(truth_counts == counts)


class CommentTests(unittest.TestCase):

    def test_comment(self):
        # test our comment stripping logic
        f = g.trimesh.util.comment_strip

        text = 'hey whats up'
        assert f(text) == text

        text = '#hey whats up'
        assert f(text) == ''

        text = '   # hey whats up '
        assert f(text) == ''

        text = '# naahah\nhey whats up'
        assert f(text) == 'hey whats up'

        text = '#naahah\nhey whats up\nhi'
        assert f(text) == 'hey whats up\nhi'

        text = '#naahah\nhey whats up\n hi'
        assert f(text) == 'hey whats up\n hi'

        text = '#naahah\nhey whats up\n hi#'
        assert f(text) == 'hey whats up\n hi'

        text = 'hey whats up# see here\n hi#'
        assert f(text) == 'hey whats up\n hi'


class ArrayToString(unittest.TestCase):
    def test_converts_an_unstructured_1d_array(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(np.array([1, 2, 3])),
            '1 2 3'
        )

    def test_converts_an_unstructured_int_array(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(np.array([[1, 2, 3], [4, 5, 6]])),
            '1 2 3\n4 5 6'
        )

    def test_converts_an_unstructured_float_array(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)
            ),
            '1.00000000 2.00000000 3.00000000\n4.00000000 5.00000000 6.00000000'
        )

    def test_uses_the_specified_column_delimiter(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([[1, 2, 3], [4, 5, 6]]), col_delim='col'),
            '1col2col3\n4col5col6'
        )

    def test_uses_the_specified_row_delimiter(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([[1, 2, 3], [4, 5, 6]]), row_delim='row'),
            '1 2 3row4 5 6'
        )

    def test_uses_the_specified_value_format(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64),
                value_format='{:.1f}'),
            '1.0 2.0 3.0\n4.0 5.0 6.0'
        )

    def test_supports_uints(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([1, 2, 3], dtype=np.uint8)),
            '1 2 3'
        )

    def test_supports_repeat_format(self):
        self.assertEqual(
            g.trimesh.util.array_to_string(
                np.array([[1, 2, 3], [4, 5, 6]]), value_format='{} {}'),
            '1 1 2 2 3 3\n4 4 5 5 6 6'
        )

    def test_raises_if_array_is_structured(self):
        with self.assertRaises(ValueError):
            g.trimesh.util.array_to_string(np.array(
                [(1, 1.1), (2, 2.2)],
                dtype=[('some_int', np.int64), ('some_float', np.float64)]
            ))

    def test_raises_if_array_is_not_flat(self):
        with self.assertRaises(ValueError):
            g.trimesh.util.array_to_string(np.array([[[1, 2, 3], [4, 5, 6]]]))


class StructuredArrayToString(unittest.TestCase):

    def test_converts_a_structured_array_with_1d_elements(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                )
            ),
            '1 1.10000000\n2 2.20000000'
        )

    def test_converts_a_structured_array_with_2d_elements(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [([1, 2], 1.1), ([3, 4], 2.2)],
                    dtype=[('some_int', np.int64, 2),
                           ('some_float', np.float64)]
                )
            ),
            '1 2 1.10000000\n3 4 2.20000000'
        )

    def test_uses_the_specified_column_delimiter(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                ),
                col_delim='col'
            ),
            '1col1.10000000\n2col2.20000000'
        )

    def test_uses_the_specified_row_delimiter(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                ),
                row_delim='row'
            ),
            '1 1.10000000row2 2.20000000'
        )

    def test_uses_the_specified_value_format(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                ),
                value_format='{:.1f}'
            ),
            '1.0 1.1\n2.0 2.2'
        )

    def test_supports_uints(self):
        self.assertEqual(
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.uint8), ('some_float', np.float64)]
                )
            ),
            '1 1.10000000\n2 2.20000000'
        )

    def test_raises_if_array_is_unstructured(self):
        with self.assertRaises(ValueError):
            g.trimesh.util.structured_array_to_string(np.ndarray([1, 2, 3]))

    def test_raises_if_value_format_specifies_repeats(self):
        with self.assertRaises(ValueError):
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [(1, 1.1), (2, 2.2)],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                ),
                value_format='{} {}'
            )

    def test_raises_if_array_is_not_flat(self):
        with self.assertRaises(ValueError):
            g.trimesh.util.structured_array_to_string(
                np.array(
                    [[(1, 1.1), (2, 2.2)], [(1, 1.1), (2, 2.2)]],
                    dtype=[('some_int', np.int64), ('some_float', np.float64)]
                )
            )


if __name__ == '__main__':
    trimesh.util.attach_to_log()
    unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class DXMLTest(g.unittest.TestCase):

    def test_abaqus(self):
        # an assembly with instancing
        s = g.get_mesh('cube1.3dxml')

        # should be 1 unique meshes
        assert len(s.geometry) == 1


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
primitives.py
----------------

Subclasses of Trimesh objects that are parameterized as primitives.

Useful because you can move boxes and spheres around
and then use trimesh operations on them at any point.
"""
import numpy as np
import abc

from . import util
from . import sample
from . import caching
from . import inertia
from . import creation
from . import triangles
from . import transformations as tf

from .base import Trimesh
from .constants import log, tol

# immutable identity matrix for checks
_IDENTITY = np.eye(4)
_IDENTITY.flags.writeable = False


class _Primitive(Trimesh):
    """
    Geometric _Primitives which are a subclass of Trimesh.
    Mesh is generated lazily when vertices or faces are requested.
    """

    # ignore superclass copy directives
    __copy__ = None
    __deepcopy__ = None

    def __init__(self, *args, **kwargs):
        super(_Primitive, self).__init__(*args, **kwargs)
        self._data.clear()
        self._validate = False
        # make sure any cached numpy arrays have
        # set `array.flags.writable = False`
        self._cache.force_immutable = True

    def __repr__(self):
        return '<trimesh.primitives.{}>'.format(
            type(self).__name__)

    @property
    def faces(self):
        stored = self._cache['faces']
        if util.is_shape(stored, (-1, 3)):
            return stored
        self._create_mesh()
        return self._cache['faces']

    @faces.setter
    def faces(self, values):
        log.warning('primitive faces are immutable: not setting!')

    @property
    def vertices(self):
        stored = self._cache['vertices']
        if util.is_shape(stored, (-1, 3)):
            return stored

        self._create_mesh()
        return self._cache['vertices']

    @vertices.setter
    def vertices(self, values):
        if values is not None:
            log.warning('primitive vertices are immutable: not setting!')

    @property
    def face_normals(self):
        # we need to avoid the logic in the superclass that
        # is specific to the data model prioritizing faces
        stored = self._cache['face_normals']
        if util.is_shape(stored, (-1, 3)):
            return stored
        # just calculate if not stored
        unit, valid = triangles.normals(self.triangles)
        normals = np.zeros((len(valid), 3))
        normals[valid] = unit
        # store and return
        self._cache['face_normals'] = normals
        return normals

    @face_normals.setter
    def face_normals(self, values):
        if values is not None:
            log.warning('Primitive face normals are immutable!')

    @property
    def transform(self):
        """
        The transform of the Primitive object.

        Returns
        -------------
        transform : (4, 4) float
          Homogeneous transformation matrix
        """
        return self.primitive.transform

    @abc.abstractmethod
    def to_dict(self):
        """
        Should be implemented by each primitive.
        """
        raise NotImplementedError()

    def copy(self, **kwargs):
        """
        Return a copy of the Primitive object.

        Returns
        -------------
        copied : object
          Copy of current primitive
        """
        # get the constructor arguments
        kwargs.update(self.to_dict())
        # remove the type indicator, i.e. `Cylinder`
        kwargs.pop('kind')
        # create a new object with kwargs
        return type(self)(**kwargs)

    def to_mesh(self):
        """
        Return a copy of the Primitive object as a Trimesh.

        Returns
        ------------
        mesh : trimesh.Trimesh
          Tessellated version of the primitive.
        """
        result = Trimesh(
            vertices=self.vertices.copy(),
            faces=self.faces.copy(),
            face_normals=self.face_normals.copy(),
            process=False)
        return result

    def apply_transform(self, matrix):
        """
        Apply a transform to the current primitive by
        applying a new transform on top of existing
        `self.primitive.transform`. If the matrix
        contains scaling it will change parameters
        like `radius` or `height` automatically.

        Parameters
        ------------
        matrix: (4, 4) float
          Homogeneous transformation
        """
        matrix = np.asanyarray(
            matrix, order='C', dtype=np.float64)
        if matrix.shape != (4, 4):
            raise ValueError('matrix must be `(4, 4)`!')
        if util.allclose(matrix, _IDENTITY, 1e-8):
            # identity matrix is a no-op
            return

        prim = self.primitive
        # copy the current transform
        current = prim.transform.copy()
        # see if matrix has scaling from the matrix
        scale = np.linalg.det(matrix[:3, :3]) ** (1.0 / 3.0)

        # the objects we handle re-scaling for
        # note that `Extrusion` is NOT supported
        kinds = (Box, Cylinder, Capsule, Sphere)
        if isinstance(self, kinds) and abs(scale - 1.0) > 1e-8:
            # scale the primitive attributes
            if hasattr(prim, 'height'):
                prim.height *= scale
            if hasattr(prim, 'radius'):
                prim.radius *= scale
            if hasattr(prim, 'extents'):
                prim.extents *= scale
            # scale the translation of the current matrix
            current[:3, 3] *= scale
            # apply new matrix, rescale, translate, current
            updated = util.multi_dot([
                matrix,
                tf.scale_matrix(1.0 / scale),
                current])
        else:
            # without scaling just multiply
            updated = np.dot(matrix, current)

        # make sure matrix is a rigid transform
        if not tf.is_rigid(updated):
            raise ValueError("Couldn't produce rigid transform!")

        # apply the new matrix
        self.primitive.transform = updated

        return self

    def _create_mesh(self):
        raise ValueError('Primitive doesn\'t define mesh creation!')


class _PrimitiveAttributes(object):
    """
    Hold the mutable data which defines a primitive.
    """

    def __init__(self, parent, defaults, kwargs):
        """
        Hold the attributes for a Primitive.

        Parameters
        ------------
        parent : _Primitive
          Parent object reference.
        defaults : dict
          The default values for this primitive type.
        kwargs : dict
          User-passed values, i.e. {'radius': 10.0}
        """
        # store actual data in parent object
        self._data = parent._data
        # default values define the keys
        self._defaults = defaults
        # store a reference to the parent ubject
        self._parent = parent
        # start with a copy of all default objects
        self._data.update(defaults)
        # store whether this data is mutable after creation
        self._mutable = kwargs.get('mutable', True)
        # assign the keys passed by the user only if
        # they are a property of this primitive
        for key, default in defaults.items():
            value = kwargs.get(key, None)
            if value is not None:
                # convert passed data into type of defaults
                self._data[key] = util.convert_like(value, default)
        # make sure stored values are immutable after setting
        if not self._mutable:
            self._data.mutable = False

    @property
    def __doc__(self):
        # this is generated dynamically as the format
        # operation can be surprisingly slow and most
        # people never call it
        import pprint
        doc = (
            'Store the attributes of a {name} object.\n\n' +
            'When these values are changed, the mesh geometry will \n' +
            'automatically be updated to reflect the new values.\n\n' +
            'Available properties and their default values are:\n {defaults}' +
            '\n\nExample\n---------------\n' +
            'p = trimesh.primitives.{name}()\n' +
            'p.primitive.radius = 10\n' +
            '\n').format(
            name=self._parent.__class__.__name__,
            defaults=pprint.pformat(
                self._defaults,
                width=-1)[1:-1])
        return doc

    def __getattr__(self, key):
        if key.startswith('_'):
            return super(_PrimitiveAttributes, self).__getattr__(key)
        elif key == 'center':
            # this whole __getattr__ is a little hacky
            return self._data['transform'][:3, 3]
        elif key in self._defaults:
            return util.convert_like(self._data[key],
                                     self._defaults[key])
        raise AttributeError(
            "primitive object has no attribute '{}' ".format(key))

    def __setattr__(self, key, value):
        if key.startswith('_'):
            return super(_PrimitiveAttributes,
                         self).__setattr__(key, value)
        elif key == 'center':
            value = np.array(value, dtype=np.float64)
            transform = np.eye(4)
            transform[:3, 3] = value
            self._data['transform'] = transform
            return
        elif key in self._defaults:
            if self._mutable:
                self._data[key] = util.convert_like(
                    value, self._defaults[key])
            else:
                raise ValueError(
                    'Primitive is configured as immutable! Cannot set attribute!')
        else:
            keys = list(self._defaults.keys())
            raise ValueError(
                'Only default attributes {} can be set!'.format(keys))

    def __dir__(self):
        result = sorted(dir(type(self)) +
                        list(self._defaults.keys()))
        return result


class Cylinder(_Primitive):

    def __init__(self, radius=1.0, height=1.0,
                 transform=None, sections=32, **kwargs):
        """
        Create a Cylinder Primitive, a subclass of Trimesh.

        Parameters
        -------------
        radius : float
          Radius of cylinder
        height : float
          Height of cylinder
        transform : (4, 4) float
          Homogeneous transformation matrix
        sections : int
          Number of facets in circle
        """
        super(Cylinder, self).__init__(**kwargs)

        defaults = {'height': 10.0,
                    'radius': 1.0,
                    'transform': np.eye(4),
                    'sections': 32}
        self.primitive = _PrimitiveAttributes(
            self,
            defaults,
            {'height': height,
             'radius': radius,
             'transform': transform,
             'sections': sections})

    @caching.cache_decorator
    def volume(self):
        """
        The analytic volume of the cylinder primitive.

        Returns
        ---------
        volume : float
          Volume of the cylinder
        """
        volume = ((np.pi * self.primitive.radius ** 2) *
                  self.primitive.height)
        return volume

    @caching.cache_decorator
    def moment_inertia(self):
        """
        The analytic inertia tensor of the cylinder primitive.

        Returns
        ----------
        tensor: (3, 3) float
          3D inertia tensor
        """

        tensor = inertia.cylinder_inertia(
            mass=self.volume,
            radius=self.primitive.radius,
            height=self.primitive.height,
            transform=self.primitive.transform)
        return tensor

    @caching.cache_decorator
    def direction(self):
        """
        The direction of the cylinder's axis.

        Returns
        --------
        axis: (3,) float, vector along the cylinder axis
        """
        axis = np.dot(self.primitive.transform, [0, 0, 1, 0])[:3]
        return axis

    @property
    def segment(self):
        """
        A line segment which if inflated by cylinder radius
        would represent the cylinder primitive.

        Returns
        -------------
        segment : (2, 3) float
          Points representing a single line segment
        """
        # half the height
        half = self.primitive.height / 2.0
        # apply the transform to the Z- aligned segment
        points = np.dot(
            self.primitive.transform,
            np.transpose([[0, 0, -half, 1],
                          [0, 0, half, 1]])).T[:, :3]
        return points

    def to_dict(self):
        """
        Get a copy of the current Cylinder primitive as
        a JSON-serializable dict that matches the schema
        in `trimesh/resources/schema/cylinder.schema.json`

        Returns
        ----------
        as_dict : dict
          Serializable data for this primitive.
        """
        return {'kind': 'cylinder',
                'transform': self.primitive.transform.tolist(),
                'radius': float(self.primitive.radius),
                'height': float(self.primitive.height), }

    def buffer(self, distance):
        """
        Return a cylinder primitive which covers the source
        cylinder by distance: radius is inflated by distance
        height by twice the distance.

        Parameters
        ------------
        distance : float
          Distance to inflate cylinder radius and height

        Returns
        -------------
        buffered : Cylinder
         Cylinder primitive inflated by distance
        """
        distance = float(distance)
        buffered = Cylinder(
            height=self.primitive.height + distance * 2,
            radius=self.primitive.radius + distance,
            transform=self.primitive.transform.copy())
        return buffered

    def _create_mesh(self):
        log.debug('creating mesh for Cylinder primitive')
        mesh = creation.cylinder(radius=self.primitive.radius,
                                 height=self.primitive.height,
                                 sections=self.primitive.sections,
                                 transform=self.primitive.transform)

        self._cache['vertices'] = mesh.vertices
        self._cache['faces'] = mesh.faces
        self._cache['face_normals'] = mesh.face_normals


class Capsule(_Primitive):

    def __init__(self,
                 radius=1.0,
                 height=10.0,
                 transform=None,
                 sections=32,
                 **kwargs):
        """
        Create a Capsule Primitive, a subclass of Trimesh.

        Parameters
        ----------
        radius : float
          Radius of cylinder
        height : float
          Height of cylinder
        transform : (4, 4) float
          Transformation matrix
        sections : int
          Number of facets in circle
        """
        super(Capsule, self).__init__(**kwargs)

        defaults = {'height': 1.0,
                    'radius': 1.0,
                    'transform': np.eye(4),
                    'sections': 32}
        self.primitive = _PrimitiveAttributes(
            self,
            defaults,
            {'height': height,
             'radius': radius,
             'transform': transform,
             'sections': sections})

    @property
    def transform(self):
        return self.primitive.transform

    def to_dict(self):
        """
        Get a copy of the current Capsule primitive as
        a JSON-serializable dict that matches the schema
        in `trimesh/resources/schema/capsule.schema.json`

        Returns
        ----------
        as_dict : dict
          Serializable data for this primitive.
        """
        return {'kind': 'capsule',
                'transform': self.primitive.transform.tolist(),
                'height': float(self.primitive.height),
                'radius': float(self.primitive.radius)}

    @caching.cache_decorator
    def direction(self):
        """
        The direction of the capsule's axis.

        Returns
        --------
        axis : (3,) float
          Vector along the cylinder axis
        """
        axis = np.dot(self.primitive.transform,
                      [0, 0, 1, 0])[:3]
        return axis

    def _create_mesh(self):
        log.debug('creating mesh for `Capsule` primitive')

        mesh = creation.capsule(radius=self.primitive.radius,
                                height=self.primitive.height)
        mesh.apply_transform(self.primitive.transform)

        self._cache['vertices'] = mesh.vertices
        self._cache['faces'] = mesh.faces
        self._cache['face_normals'] = mesh.face_normals


class Sphere(_Primitive):

    def __init__(self,
                 radius=1.0,
                 center=None,
                 transform=None,
                 subdivisions=3,
                 **kwargs):
        """
        Create a Sphere Primitive, a subclass of Trimesh.

        Parameters
        ----------
        radius : float
          Radius of sphere
        center : None or (3,) float
          Center of sphere.
        transform : None or (4, 4) float
          Full homogeneous transform. Pass `center` OR `transform.
        subdivisions : int
          Number of subdivisions for icosphere.
        """

        super(Sphere, self).__init__(**kwargs)

        defaults = {'radius': 1.0,
                    'transform': np.eye(4),
                    'subdivisions': 3}
        constructor = {'radius': float(radius),
                       'subdivisions': int(subdivisions)}
        # center is a helper method for "transform"
        # since a sphere is rotationally symmetric
        if center is not None:
            if transform is not None:
                raise ValueError(
                    'only one of `center` and `transform` may be passed!')
            translate = np.eye(4)
            translate[:3, 3] = center
            constructor['transform'] = translate
        elif transform is not None:
            constructor['transform'] = transform

        # create the attributes object
        self.primitive = _PrimitiveAttributes(
            self, defaults, constructor)

    @property
    def center(self):
        return self.primitive.center

    @center.setter
    def center(self, value):
        self.primitive.center = value

    def to_dict(self):
        """
        Get a copy of the current Sphere primitive as
        a JSON-serializable dict that matches the schema
        in `trimesh/resources/schema/sphere.schema.json`

        Returns
        ----------
        as_dict : dict
          Serializable data for this primitive.
        """
        return {'kind': 'sphere',
                'transform': self.primitive.transform.tolist(),
                'radius': float(self.primitive.radius)}

    @property
    def bounds(self):
        # no docstring so will inherit Trimesh docstring
        # return exact bounds from primitive center and radius (rather than faces)
        # self.extents will also use this information
        bounds = np.array([self.primitive.center - self.primitive.radius,
                           self.primitive.center + self.primitive.radius])
        return bounds

    @property
    def bounding_box_oriented(self):
        # for a sphere the oriented bounding box is the same as the axis aligned
        # bounding box, and a sphere is the absolute slowest case for the OBB calculation
        # as it is a convex surface with a ton of face normals that all need to
        # be checked
        return self.bounding_box

    @caching.cache_decorator
    def area(self):
        """
        Surface area of the current sphere primitive.

        Returns
        --------
        area: float, surface area of the sphere Primitive
        """

        area = 4.0 * np.pi * (self.primitive.radius ** 2)
        return area

    @caching.cache_decorator
    def volume(self):
        """
        Volume of the current sphere primitive.

        Returns
        --------
        volume: float, volume of the sphere Primitive
        """

        volume = (4.0 * np.pi * (self.primitive.radius ** 3)) / 3.0
        return volume

    @caching.cache_decorator
    def moment_inertia(self):
        """
        The analytic inertia tensor of the sphere primitive.

        Returns
        ----------
        tensor: (3, 3) float
          3D inertia tensor.
        """
        return inertia.sphere_inertia(
            mass=self.volume,
            radius=self.primitive.radius)

    def _create_mesh(self):
        log.debug('creating mesh for Sphere primitive')
        unit = creation.icosphere(subdivisions=self.primitive.subdivisions)
        unit.vertices *= self.primitive.radius
        unit.vertices += self.primitive.center

        self._cache['vertices'] = unit.vertices
        self._cache['faces'] = unit.faces
        self._cache['face_normals'] = unit.face_normals


class Box(_Primitive):

    def __init__(self,
                 extents=None,
                 transform=None,
                 **kwargs):
        """
        Create a Box Primitive as a subclass of Trimesh

        Parameters
        ----------
        extents : (3,) float
          Length of each side of the 3D box
        transform : (4, 4) float
          Homogeneous transformation matrix for box center
        """
        super(Box, self).__init__(**kwargs)
        defaults = {'transform': np.eye(4),
                    'extents': np.ones(3)}
        self.primitive = _PrimitiveAttributes(
            self,
            defaults,
            {'extents': extents,
             'transform': transform})

    def to_dict(self):
        """
        Get a copy of the current Box primitive as
        a JSON-serializable dict that matches the schema
        in `trimesh/resources/schema/box.schema.json`

        Returns
        ----------
        as_dict : dict
          Serializable data for this primitive.
        """
        return {'kind': 'box',
                'transform': self.primitive.transform.tolist(),
                'extents': self.primitive.extents.tolist()}

    @property
    def transform(self):
        return self.primitive.transform

    def sample_volume(self, count):
        """
        Return random samples from inside the volume of the box.

        Parameters
        -------------
        count : int
          Number of samples to return

        Returns
        ----------
        samples : (count, 3) float
          Points inside the volume
        """
        samples = sample.volume_rectangular(
            extents=self.primitive.extents,
            count=count,
            transform=self.primitive.transform)
        return samples

    def sample_grid(self, count=None, step=None):
        """
        Return a 3D grid which is contained by the box.
        Samples are either 'step' distance apart, or there are
        'count' samples per box side.

        Parameters
        -----------
        count : int or (3,) int
          If specified samples are spaced with np.linspace
        step : float or (3,) float
          If specified samples are spaced with np.arange

        Returns
        -----------
        grid : (n, 3) float
          Points inside the box
        """

        if (count is not None and
                step is not None):
            raise ValueError('only step OR count can be specified!')

        # create pre- transform bounds from extents
        bounds = np.array([-self.primitive.extents,
                           self.primitive.extents]) * .5

        if step is not None:
            grid = util.grid_arange(bounds, step=step)
        elif count is not None:
            grid = util.grid_linspace(bounds, count=count)
        else:
            raise ValueError('either count or step must be specified!')

        transformed = tf.transform_points(
            grid, matrix=self.primitive.transform)
        return transformed

    @property
    def is_oriented(self):
        """
        Returns whether or not the current box is rotated at all.
        """
        if util.is_shape(self.primitive.transform, (4, 4)):
            return not np.allclose(self.primitive.transform[
                                   0:3, 0:3], np.eye(3))
        else:
            return False

    @caching.cache_decorator
    def volume(self):
        """
        Volume of the box Primitive.

        Returns
        --------
        volume : float
          Volume of box.
        """
        volume = float(np.product(self.primitive.extents))
        return volume

    def _create_mesh(self):
        log.debug('creating mesh for Box primitive')
        box = creation.box(extents=self.primitive.extents,
                           transform=self.primitive.transform)

        self._cache['vertices'] = box.vertices
        self._cache['faces'] = box.faces
        self._cache['face_normals'] = box.face_normals

    def as_outline(self):
        """
        Return a Path3D containing the outline of the box.

        Returns
        -----------
        outline : trimesh.path.Path3D
          Outline of box primitive
        """
        # do the import in function to keep soft dependency
        from .path.creation import box_outline
        # return outline with same size as primitive
        return box_outline(
            extents=self.primitive.extents,
            transform=self.primitive.transform)


class Extrusion(_Primitive):
    def __init__(self,
                 polygon=None,
                 transform=None,
                 height=1.0,
                 **kwargs):
        """
        Create an Extrusion primitive, which
        is a subclass of Trimesh.

        Parameters
        ----------
        polygon : shapely.geometry.Polygon
          Polygon to extrude
        transform : (4, 4) float
          Transform to apply after extrusion
        height : float
          Height to extrude polygon by
        """
        # do the import here, fail early if Shapely isn't installed
        from shapely.geometry import Point
        super(Extrusion, self).__init__(**kwargs)
        # set default values
        defaults = {'polygon': Point([0, 0]).buffer(1.0),
                    'transform': np.eye(4),
                    'height': 1.0}
        self.primitive = _PrimitiveAttributes(
            self, defaults, {'transform': transform,
                             'polygon': polygon,
                             'height': height})

    @caching.cache_decorator
    def area(self):
        """
        The surface area of the primitive extrusion.

        Calculated from polygon and height to avoid mesh creation.

        Returns
        ----------
        area: float
          Surface area of 3D extrusion
        """
        # area of the sides of the extrusion
        area = abs(self.primitive.height *
                   self.primitive.polygon.length)
        # area of the two caps of the extrusion
        area += self.primitive.polygon.area * 2
        return area

    @caching.cache_decorator
    def volume(self):
        """
        The volume of the Extrusion primitive.
        Calculated from polygon and height to avoid mesh creation.

        Returns
        ----------
        volume : float
          Volume of 3D extrusion
        """
        # height may be negative
        volume = abs(self.primitive.polygon.area *
                     self.primitive.height)
        return volume

    @caching.cache_decorator
    def direction(self):
        """
        Based on the extrudes transform what is the
        vector along which the polygon will be extruded.

        Returns
        ---------
        direction : (3,) float
          Unit direction vector
        """
        # only consider rotation and signed height
        direction = np.dot(
            self.primitive.transform[:3, :3],
            [0.0, 0.0, np.sign(self.primitive.height)])
        return direction

    @property
    def origin(self):
        """
        Based on the extrude transform what is the
        origin of the plane it is extruded from.

        Returns
        -----------
        origin : (3,) float
          Origin of extrusion plane
        """
        return self.primitive.transform[:3, 3]

    @property
    def transform(self):
        return self.primitive.transform

    @caching.cache_decorator
    def bounding_box_oriented(self):
        # no docstring for inheritance
        # calculate OBB using 2D polygon and known axis
        from . import bounds
        # find the 2D bounding box using the polygon
        to_origin, box = bounds.oriented_bounds_2D(
            self.primitive.polygon.exterior.coords)
        #  3D extents
        extents = np.append(box, abs(self.primitive.height))
        # calculate to_3D transform from 2D obb
        rotation_Z = np.linalg.inv(tf.planar_matrix_to_3D(to_origin))
        rotation_Z[2, 3] = self.primitive.height / 2.0
        # combine the 2D OBB transformation with the 2D projection transform
        to_3D = np.dot(self.primitive.transform, rotation_Z)
        obb = Box(transform=to_3D,
                  extents=extents,
                  mutable=False)
        return obb

    def slide(self, distance):
        """
        Alter the transform of the current extrusion to slide it
        along its extrude_direction vector

        Parameters
        -----------
        distance : float
          Distance along self.extrude_direction to move
        """
        distance = float(distance)
        translation = np.eye(4)
        translation[2, 3] = distance
        new_transform = np.dot(self.primitive.transform.copy(),
                               translation.copy())
        self.primitive.transform = new_transform

    def buffer(self, distance, distance_height=None, **kwargs):
        """
        Return a new Extrusion object which is expanded in profile
        and in height by a specified distance.

        Parameters
        --------------
        distance : float
          Distance to buffer polygon
        distance_height : float
          Distance to buffer above and below extrusion
        kwargs : dict
          Passed to Extrusion constructor

        Returns
        ----------
        buffered : primitives.Extrusion
          Extrusion object with new values
        """
        distance = float(distance)
        # if not specified use same distance for everything
        if distance_height is None:
            distance_height = distance

        # start with current height
        height = self.primitive.height
        # if current height is negative offset by negative amount
        height += np.sign(height) * 2.0 * distance_height

        # create a new extrusion with a buffered polygon
        # use type(self) vs Extrusion to handle subclasses
        buffered = type(self)(
            transform=self.primitive.transform.copy(),
            polygon=self.primitive.polygon.buffer(distance),
            height=height,
            **kwargs)

        # slide the stock along the axis
        buffered.slide(-np.sign(height) * distance_height)

        return buffered

    def to_dict(self):
        """
        Get a copy of the current Extrusion primitive as
        a JSON-serializable dict that matches the schema
        in `trimesh/resources/schema/extrusion.schema.json`

        Returns
        ----------
        as_dict : dict
          Serializable data for this primitive.
        """
        return {'kind': 'extrusion',
                'polygon': self.primitive.polygon.wkt,
                'transform': self.primitive.transform.tolist(),
                'height': float(self.primitive.height)}

    def _create_mesh(self):
        log.debug('creating mesh for Extrusion primitive')
        # extrude the polygon along Z
        mesh = creation.extrude_polygon(
            polygon=self.primitive.polygon,
            height=self.primitive.height,
            transform=self.primitive.transform)

        # check volume here in unit tests
        if tol.strict and mesh.volume < 0.0:
            raise ValueError('matrix inverted mesh!')

        # cache mesh geometry in the primitive
        self._cache['vertices'] = mesh.vertices
        self._cache['faces'] = mesh.faces



try:
    from . import generic as g
except BaseException:
    import generic as g


class ConvexTest(g.unittest.TestCase):

    def test_convex(self):

        # store (true is_convex, mesh) tuples
        meshes = [(False, g.get_mesh('featuretype.STL')),
                  (False, g.get_mesh('quadknot.obj')),
                  (True, g.get_mesh('unit_cube.STL')),
                  (False, g.get_mesh('1002_tray_bottom.STL')),
                  (True, g.trimesh.creation.icosphere()),
                  (True, g.trimesh.creation.uv_sphere()),
                  (True, g.trimesh.creation.box()),
                  (True, g.trimesh.creation.cylinder(
                      radius=1, height=10)),
                  (True, g.trimesh.creation.capsule()),
                  (False, (g.trimesh.creation.box(extents=(1, 1, 1)) +
                           g.trimesh.creation.box(bounds=[[10, 10, 10], [12, 12, 12]])))]

        for is_convex, mesh in meshes:
            assert mesh.is_watertight

            # mesh should match it's true convexity
            assert mesh.is_convex == is_convex

            hulls = []
            volume = []

            # transform the mesh around space and check volume
            # use repeatable transforms to avoid spurious failures
            for T in g.transforms[::2]:
                permutated = mesh.copy()
                permutated.apply_transform(T)
                hulls.append(permutated.convex_hull)
                volume.append(hulls[-1].volume)

            # which of the volumes are close to the median volume
            close = g.np.isclose(volume,
                                 g.np.median(volume),
                                 atol=mesh.bounding_box.volume / 1000)

            if g.platform.system() == 'Linux':
                # on linux the convex hulls are pretty robust
                close_ok = close.all()
            else:
                # on windows sometimes there is flaky numerical weirdness
                # which causes spurious CI failures, so use softer metric
                # for success: here of 90% of values are close to the median
                # then declare everything hunky dory
                ratio = close.sum() / float(len(close))
                close_ok = ratio > .9

            if not close_ok:
                g.log.error('volume inconsistent: {}'.format(volume))
                raise ValueError('volume is inconsistent on {}'.format(
                    mesh.metadata['file_name']))
            assert min(volume) > 0.0

            if not all(i.is_winding_consistent for i in hulls):
                raise ValueError(
                    'mesh %s reported bad winding on convex hull!',
                    mesh.metadata['file_name'])

            if not all(i.is_convex for i in hulls):
                raise ValueError('mesh %s reported non-convex convex hull!',
                                 mesh.metadata['file_name'])

    def test_primitives(self):
        for prim in [g.trimesh.primitives.Sphere(),
                     g.trimesh.primitives.Cylinder(),
                     g.trimesh.primitives.Box()]:
            assert prim.is_convex
            # convex things should have hulls of the same volume
            # convert to mesh to get tessellated volume rather than
            # analytic primitive volume
            tess = prim.to_mesh()
            assert g.np.isclose(tess.convex_hull.volume,
                                tess.volume)

    def test_projections(self):
        # check the vertex projection onto adjacent face plane
        # this is used to calculate convexity
        for m in g.get_meshes(4):
            assert (len(m.face_adjacency_projections) ==
                    (len(m.face_adjacency)))

    def test_truth(self):
        # check a non-watertight mesh
        m = g.get_mesh('not_convex.obj')
        assert not m.is_convex


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class MedialTests(g.unittest.TestCase):

    def test_medial(self):
        p = g.get_mesh('2D/wrench.dxf')
        assert p.is_closed

        assert len(p.polygons_full) == 1
        poly = p.polygons_full[0]
        medial = p.medial_axis()
        points = medial.vertices.view(g.np.ndarray)
        assert all(poly.contains(g.Point(v))
                   for v in points)

        # circles are a special case for medial axis
        poly = g.Point([0, 0]).buffer(1.0)
        # construct a Path2D from the polygon medial axis
        med = g.trimesh.path.Path2D(
            **g.trimesh.path.exchange.misc.edges_to_path(
                *g.trimesh.path.polygons.medial_axis(poly)))
        # should have returned a single tiny line
        # with midpoint at origin
        assert len(med.vertices) == 2
        assert len(med.entities) == 1
        assert float(med.vertices.mean(axis=0).ptp()) < 1e-8


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
units.py
--------------
Deal with physical unit systems (i.e. inches, mm)

Very basic conversions, and no requirement for
sympy.physics.units or pint.
"""
from .constants import log
from . import resources


def unit_conversion(current, desired):
    """
    Calculate the conversion from one set of units to another.

    Parameters
    ---------
    current : str
        Unit system values are in now (eg 'millimeters')
    desired : str
        Unit system we'd like values in (eg 'inches')

    Returns
    ---------
    conversion : float
        Number to multiply by to put values into desired units
    """
    # scaling factors from various unit systems to inches
    to_inch = resources.get(
        'units_to_inches.json', decode_json=True)

    current = str(current).strip().lower()
    desired = str(desired).strip().lower()
    conversion = to_inch[current] / to_inch[desired]
    return conversion


def units_from_metadata(obj, guess=True):
    """
    Try to extract hints from metadata and if that fails
    guess based on the object scale.


    Parameters
    ------------
    obj: object
        Has attributes 'metadata' (dict) and 'scale' (float)
    guess : bool
        If metadata doesn't indicate units, guess from scale

    Returns
    ------------
    units: str
        A guess of what the units might be
    """
    to_inch = resources.get(
        'units_to_inches.json', decode_json=True)

    # try to guess from metadata
    for key in ['file_name', 'name']:
        if key not in obj.metadata:
            continue
        # get the string which might contain unit hints
        hints = obj.metadata[key].lower()
        if 'unit' in hints:
            # replace all delimiter options with white space
            for delim in '_-.':
                hints = hints.replace(delim, ' ')
            # loop through each hint
            for hint in hints.strip().split():
                # key word is "unit" or "units"
                if 'unit' not in hint:
                    continue
                # get rid of keyword and whitespace
                hint = hint.replace(
                    'units', '').replace(
                        'unit', '').strip()
                # if the hint is a valid unit return it
                if hint in to_inch:
                    return hint

    if not guess:
        raise ValueError('no units and not allowed to guess')

    # we made it to the wild ass guess section
    # if the scale is larger than 100 mystery units
    # declare the model to be millimeters, otherwise inches
    log.debug('no units: guessing from scale')
    if float(obj.scale) > 100.0:
        return 'millimeters'
    else:
        return 'inches'


def _convert_units(obj, desired, guess=False):
    """
    Given an object with scale and units try to scale
    to different units via the object's `apply_scale`.

    Parameters
    ---------
    obj :  object
        With apply_scale method (i.e. Trimesh, Path2D, etc)
    desired : str
        Units desired (eg 'inches')
    guess:   bool
        Whether we are allowed to guess the units
        if they are not specified.
    """
    if obj.units is None:
        # try to extract units from metadata
        # if nothing specified in metadata and not allowed
        # to guess will raise a ValueError
        obj.units = units_from_metadata(obj, guess=guess)

    log.debug('converting units from %s to %s', obj.units, desired)
    # float, conversion factor
    conversion = unit_conversion(obj.units, desired)

    # apply scale uses transforms which preserve
    # cached properties rather than just multiplying vertices
    obj.apply_scale(conversion)
    # units are now desired units
    obj.units = desired



try:
    from . import generic as g
except BaseException:
    import generic as g


class ViewTest(g.unittest.TestCase):

    def test_JSHTML(self):

        import trimesh.viewer.notebook as js

        m = g.get_mesh('featuretype.STL')
        s = m.scene()

        # a viewable scene
        html = js.scene_to_html(s)
        assert len(html) > 0

        # check it out as an LXML document
        from lxml.html import fromstring
        h = fromstring(html)

        # should have some children
        children = list(h.body.iterchildren())
        assert len(children) >= 2

    def test_inNB(self):
        import trimesh.viewer.notebook as js

        # should only return True inside
        # jypyter/IPython notebooks
        assert not js.in_notebook()


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

import numpy as np


class CameraTests(g.unittest.TestCase):

    def test_K(self):
        resolution = (320, 240)
        fov = (60, 40)
        camera = g.trimesh.scene.Camera(
            resolution=resolution,
            fov=fov)

        # ground truth matrix
        K_expected = np.array([[277.128, 0, 160],
                               [0, 329.697, 120],
                               [0, 0, 1]],
                              dtype=np.float64)

        assert np.allclose(camera.K, K_expected, rtol=1e-3)

        # check to make sure assignment from matrix works
        K_set = K_expected.copy()
        K_set[:2, 2] = 300
        camera.K = K_set
        assert np.allclose(camera.resolution, 600)

    def test_consistency(self):
        resolution = (320, 240)
        focal = None
        fov = (60, 40)
        camera = g.trimesh.scene.Camera(
            resolution=resolution,
            focal=focal,
            fov=fov)
        assert np.allclose(camera.fov, fov)
        camera = g.trimesh.scene.Camera(
            resolution=resolution,
            focal=camera.focal,
            fov=None)
        assert np.allclose(camera.fov, fov)

    def test_focal_updates_on_resolution_change(self):
        """
        Test changing resolution with set fov updates focal.
        """
        base_res = (320, 240)
        updated_res = (640, 480)
        fov = (60, 40)

        # start with initial data
        base_cam = g.trimesh.scene.Camera(
            resolution=base_res,
            fov=fov)
        # update both focal length and resolution
        base_focal = base_cam.focal
        base_cam.resolution = updated_res

        assert not g.np.allclose(base_cam.focal,
                                 base_focal)

        # camera created with same arguments should
        # have the same values
        new_cam = g.trimesh.scene.Camera(
            resolution=updated_res,
            fov=fov)
        assert g.np.allclose(base_cam.focal, new_cam.focal)

    def test_fov_updates_on_resolution_change(self):
        """
        Test changing resolution with set focal updates fov.
        """
        base_res = (320, 240)
        updated_res = (640, 480)
        focal = (100, 100)
        base_cam = g.trimesh.scene.Camera(
            resolution=base_res,
            focal=focal)
        base_fov = base_cam.fov
        base_cam.resolution = updated_res
        assert base_cam.fov is not base_fov
        new_cam = g.trimesh.scene.Camera(
            resolution=updated_res,
            focal=focal,
        )
        np.testing.assert_allclose(base_cam.fov, new_cam.fov)

    def test_lookat(self):
        """
        Test the "look at points" function
        """
        # original points
        ori = np.array([[-1, -1],
                        [1, -1],
                        [1, 1],
                        [-1, 1]])

        for _i in range(10):
            # set the extents to be random but positive
            extents = g.random() * 10
            points = g.trimesh.util.stack_3D(ori.copy() * extents)

            fov = g.np.array([20, 50])

            # offset the points by a random amount
            offset = (g.random(3) - 0.5) * 100
            T = g.trimesh.scene.cameras.look_at(points + offset, fov)

            # check using trig
            check = (points.ptp(axis=0)[:2] / 2.0) / \
                g.np.tan(np.radians(fov / 2))
            check += points[:, 2].mean()

            # Z should be the same as maximum trig option
            assert np.linalg.inv(T)[2, 3] >= check.max()

        # just run to test other arguments
        # TODO(unknown): find the way to test it correctly
        g.trimesh.scene.cameras.look_at(points, fov, center=points[0])
        g.trimesh.scene.cameras.look_at(points, fov, distance=1)

    def test_ray_index(self):
        # make sure to_rays is giving valid indexes
        s = g.trimesh.scene.Scene()
        res = g.np.array([512, 512])
        for i in range(0, 1000, 79):
            current = res + i
            s.camera.resolution = current
            # get ray index of camera
            rid = s.camera.to_rays()[1]
            assert all(rid.min(axis=0) == 0)
            assert all(rid.max(axis=0) == current - 1)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

import io


class ExportTest(g.unittest.TestCase):

    def test_export(self):

        from trimesh.exceptions import ExceptionWrapper

        export_types = set(k for k, v in
                           g.trimesh.exchange.export._mesh_exporters.items()
                           if not isinstance(v, ExceptionWrapper))

        meshes = list(g.get_meshes(8))
        # make sure we've got something with texture
        meshes.append(g.get_mesh('fuze.obj'))

        for mesh in meshes:
            # disregard texture
            mesh.merge_vertices(merge_tex=True, merge_norm=True)
            for file_type in export_types:
                # skip pointcloud format
                if file_type in ['xyz', 'gltf']:
                    # a pointcloud format
                    continue
                # run the export
                export = mesh.export(file_type=file_type)
                # if nothing returned log the message
                if export is None or len(export) == 0:
                    raise ValueError(
                        'No data exported %s to %s',
                        mesh.metadata['file_name'],
                        file_type)

                if file_type in [
                        'dae',     # collada, no native importers
                        'collada',  # collada, no native importers
                        'msgpack',  # kind of flaky, but usually works
                        'drc']:  # DRC is not a lossless format
                    g.log.warning(
                        'no native loaders implemented for collada!')
                    continue

                g.log.info('Export/import testing on %s',
                           mesh.metadata['file_name'])

                # if export is string or bytes wrap as pseudo file object
                if isinstance(export, str) or isinstance(export, bytes):
                    file_obj = g.io_wrap(export)
                else:
                    file_obj = export

                loaded = g.trimesh.load(file_obj=file_obj,
                                        file_type=file_type,
                                        process=True,
                                        merge_norm=True,
                                        merge_tex=True)

                # if we exported as GLTF/dae it will come back as a Scene
                if isinstance(loaded, g.trimesh.Scene) and isinstance(
                        mesh, g.trimesh.Trimesh):
                    assert len(loaded.geometry) == 1
                    loaded = next(iter(loaded.geometry.values()))

                if (not g.trimesh.util.is_shape(loaded._data['faces'], (-1, 3)) or
                    not g.trimesh.util.is_shape(loaded._data['vertices'], (-1, 3)) or
                        loaded.faces.shape != mesh.faces.shape):
                    g.log.error('Export -> import for %s on %s wrong shape!',
                                file_type,
                                mesh.metadata['file_name'])

                if loaded.vertices is None:
                    g.log.error('Export -> import for %s on %s gave None for vertices!',
                                file_type,
                                mesh.metadata['file_name'])

                if loaded.faces.shape != mesh.faces.shape:
                    raise ValueError('export cycle {} on {} gave faces {}->{}!'.format(
                        file_type,
                        mesh.metadata['file_name'],
                        str(mesh.faces.shape),
                        str(loaded.faces.shape)))

                if loaded.vertices.shape != mesh.vertices.shape:
                    raise ValueError('export cycle {} on {} gave vertices {}->{}!'.format(
                        file_type,
                        mesh.metadata['file_name'],
                        mesh.vertices.shape,
                        loaded.vertices.shape))

                # try exporting/importing certain file types by name
                if file_type in ['obj', 'stl', 'ply', 'off']:
                    temp = g.tempfile.NamedTemporaryFile(suffix='.' + file_type,
                                                         delete=False)
                    # windows throws permissions errors if you keep it open
                    temp.close()

                    mesh.export(temp.name)
                    load = g.trimesh.load(temp.name)
                    # manual cleanup
                    g.os.remove(temp.name)

                    assert mesh.faces.shape == load.faces.shape
                    assert mesh.vertices.shape == load.vertices.shape

            # if we're not on linux don't run meshlab tests
            if not g.is_linux:
                continue
            # formats exportable by trimesh and importable by meshlab
            # make sure things we export can be loaded by meshlab
            both = set(g.meshlab_formats).intersection(
                set(export_types))

            # additional options to pass to exporters to try to ferret
            # out combinations which lead to invalid output
            kwargs = {'ply': [{'vertex_normal': True,
                               'encoding': 'ascii'},
                              {'vertex_normal': True,
                               'encoding': 'binary'},
                              {'vertex_normal': False,
                               'encoding': 'ascii'},
                              {'vertex_normal': False,
                               'encoding': 'binary'}],
                      'stl': [{'file_type': 'stl'},
                              {'file_type': 'stl_ascii'}]}

            # make sure input mesh has garbage removed
            mesh._validate = True
            # since we're going to be looking for exact export
            # counts remove anything small/degenerate again
            mesh.process()

            # run through file types supported by both meshlab and trimesh
            for file_type in both:
                # pull different exporter options for the format
                if file_type in kwargs:
                    options = kwargs[file_type]
                else:
                    options = [{}]

                # try each combination of options
                for option in options:
                    temp = g.tempfile.NamedTemporaryFile(
                        suffix='.' + file_type,
                        delete=False)
                    temp_off = g.tempfile.NamedTemporaryFile(
                        suffix='.off',
                        delete=False)
                    # windows throws permissions errors if you keep it open
                    temp.close()
                    temp_off.close()
                    # write over the tempfile
                    option['file_obj'] = temp.name
                    mesh.export(**option)

                    # will raise CalledProcessError if meshlab
                    # can't successfully import the file
                    try:
                        # have meshlab take the export and convert it into
                        # an OFF file, which is basically the simplest format
                        # that uses by- reference vertices
                        # meshlabserver requires X so wrap it with XVFB
                        cmd = 'xvfb-run -a -s "-screen 0 800x600x24" meshlabserver '
                        cmd += '-i {} -o {}'.format(temp.name, temp_off.name)
                        g.subprocess.check_call(cmd, shell=True)
                    except g.subprocess.CalledProcessError as E:
                        # log the options that produced the failure
                        g.log.error('failed to export {}'.format(
                            option))
                        # raise the error again
                        raise E

                    # load meshlabs export back into trimesh
                    r = g.trimesh.load(temp_off.name)

                    # we should have the same number of vertices and faces
                    assert len(r.vertices) == len(mesh.vertices)
                    assert len(r.faces) == len(mesh.faces)

                    # manual cleanup
                    g.os.remove(temp.name)
                    g.os.remove(temp_off.name)

    def test_obj(self):
        m = g.get_mesh('textured_tetrahedron.obj', process=False)
        export = m.export(file_type='obj')
        reconstructed = g.wrapload(export, file_type='obj', process=False)
        # test that we get at least the same number of normals and texcoords out;
        # the loader may reorder vertices, so we shouldn't check direct
        # equality
        assert m.vertex_normals.shape == reconstructed.vertex_normals.shape

    def test_obj_order(self):
        """
        Make sure simple round trips through Wavefront don't
        reorder vertices.
        """
        # get a writeable temp file location
        temp = g.tempfile.NamedTemporaryFile(
            suffix='.obj',
            delete=False)
        temp.close()

        # simple solid
        x = g.trimesh.creation.icosahedron()
        x.export(temp.name)
        y = g.trimesh.load_mesh(temp.name, process=False)

        # vertices should be same shape and order
        assert g.np.allclose(x.vertices, y.vertices)
        # faces should be same
        assert g.np.allclose(x.faces, y.faces)

    def test_dict(self):
        mesh = g.get_mesh('machinist.XAML')
        assert mesh.visual.kind == 'face'
        mesh.visual.vertex_colors = mesh.visual.vertex_colors
        assert mesh.visual.kind == 'vertex'

        as_dict = mesh.to_dict()
        back = g.trimesh.Trimesh(**as_dict)  # NOQA

    def test_scene(self):
        # get a multi- mesh scene with a transform tree
        source = g.get_mesh('cycloidal.3DXML')
        # add a transform to zero scene before exporting
        source.rezero()
        # export the file as a binary GLTF file, GLB
        export = source.export(file_type='glb')

        # re- load the file as a trimesh.Scene object again
        loaded = g.wrapload(export, file_type='glb')

        # the scene should be identical after export-> import cycle
        assert g.np.allclose(loaded.extents / source.extents,
                             1.0)

    def test_gltf_path(self):
        """
        Check to make sure GLTF exports of Path2D and Path3D
        objects don't immediately crash.
        """
        path2D = g.get_mesh('2D/wrench.dxf')
        path3D = path2D.to_3D()

        a = g.trimesh.Scene(path2D).export(file_type='glb')
        b = g.trimesh.Scene(path3D).export(file_type='glb')

        assert len(a) > 0
        assert len(b) > 0

    def test_parse_file_args(self):
        """
        Test the magical trimesh.exchange.load.parse_file_args
        """
        # it's wordy
        f = g.trimesh.exchange.load.parse_file_args

        RET_COUNT = 5

        # a path that doesn't exist
        nonexists = '/banana{}'.format(g.random())
        assert not g.os.path.exists(nonexists)

        # loadable OBJ model
        exists = g.os.path.join(g.dir_models, 'tube.obj')
        assert g.os.path.exists(exists)

        # should be able to extract type from passed filename
        args = f(file_obj=exists, file_type=None)
        assert len(args) == RET_COUNT
        assert args[1] == 'obj'

        # should be able to extract correct type from longer name
        args = f(file_obj=exists, file_type='YOYOMA.oBj')
        assert len(args) == RET_COUNT
        assert args[1] == 'obj'

        # with a nonexistent file and no extension it should raise
        try:
            args = f(file_obj=nonexists, file_type=None)
        except ValueError as E:
            assert 'not a file' in str(E)
        else:
            raise ValueError('should have raised exception!')

        # nonexistent file with extension passed should return
        # file name anyway, maybe something else can handle it
        args = f(file_obj=nonexists, file_type='.ObJ')
        assert len(args) == RET_COUNT
        # should have cleaned up case
        assert args[1] == 'obj'

        # make sure overriding type works for string filenames
        args = f(file_obj=exists, file_type='STL')
        assert len(args) == RET_COUNT
        # should have used manually passed type over .obj
        assert args[1] == 'stl'

    def test_buffered_random(self):
        """Test writing to non-standard file
        """
        mesh = list(g.get_meshes(1))[0]
        with io.BufferedRandom(io.BytesIO()) as rw:
            mesh.export(rw, 'STL')
            rw.seek(0)
            binary_stl = rw.read()
            self.assertLess(0, len(binary_stl))


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class PlyTest(g.unittest.TestCase):

    def test_ply_dtype(self):
        # make sure all ply dtype strings are valid dtypes
        dtypes = g.trimesh.exchange.ply._dtypes
        for d in dtypes.values():
            # will raise if dtype string not valid
            g.np.dtype(d)

    def test_ply(self):
        m = g.get_mesh('machinist.XAML')

        assert m.visual.kind == 'face'
        assert m.visual.face_colors.ptp(axis=0).max() > 0

        export = m.export(file_type='ply')
        reconstructed = g.wrapload(export, file_type='ply')

        assert reconstructed.visual.kind == 'face'

        assert g.np.allclose(reconstructed.visual.face_colors,
                             m.visual.face_colors)

        m = g.get_mesh('reference.ply')

        assert m.visual.kind == 'vertex'
        assert m.visual.vertex_colors.ptp(axis=0).max() > 0

        export = m.export(file_type='ply')
        reconstructed = g.wrapload(export, file_type='ply')
        assert reconstructed.visual.kind == 'vertex'

        assert g.np.allclose(reconstructed.visual.vertex_colors,
                             m.visual.vertex_colors)

    def test_points(self):
        # Test reading point clouds from PLY files

        m = g.get_mesh('points_ascii.ply')
        assert isinstance(m, g.trimesh.PointCloud)
        assert m.vertices.shape == (5, 3)

        m = g.get_mesh('points_bin.ply')
        assert m.vertices.shape == (5, 3)
        assert isinstance(m, g.trimesh.PointCloud)

        m = g.get_mesh('points_emptyface.ply')
        assert m.vertices.shape == (1024, 3)
        assert isinstance(m, g.trimesh.PointCloud)

    def test_list_properties(self):
        """
        Test reading point clouds with the following metadata:
        - lists of differing length
        - multiple list properties
        - single-element properties that come after list properties
        """
        m = g.get_mesh('points_ascii_with_lists.ply')

        point_list = m.metadata['_ply_raw']['point_list']['data']
        assert g.np.array_equal(
            point_list['point_indices1'][0], g.np.array([10, 11, 12], dtype=g.np.uint32))
        assert g.np.array_equal(
            point_list['point_indices1'][1], g.np.array([10, 11], dtype=g.np.uint32))
        assert g.np.array_equal(
            point_list['point_indices2'][0], g.np.array([13, 14], dtype=g.np.uint32))
        assert g.np.array_equal(
            point_list['point_indices2'][1], g.np.array([12, 13, 14], dtype=g.np.uint32))
        assert g.np.array_equal(
            point_list['some_float'], g.np.array([1.1, 2.2], dtype=g.np.float32))

    def test_vertex_attributes(self):
        """
        Test writing vertex attributes to a ply, by reading them back and asserting the
        written attributes array matches
        """

        m = g.get_mesh('box.STL')
        test_1d_attribute = g.np.copy(m.vertices[:, 0])
        test_nd_attribute = g.np.copy(m.vertices)
        m.vertex_attributes['test_1d_attribute'] = test_1d_attribute
        m.vertex_attributes['test_nd_attribute'] = test_nd_attribute

        export = m.export(file_type='ply')
        reconstructed = g.wrapload(export,
                                   file_type='ply')

        vertex_attributes = reconstructed.metadata['_ply_raw']['vertex']['data']
        result_1d = vertex_attributes['test_1d_attribute']
        result_nd = vertex_attributes['test_nd_attribute']['f1']

        g.np.testing.assert_almost_equal(result_1d, test_1d_attribute)
        g.np.testing.assert_almost_equal(result_nd, test_nd_attribute)

    def test_face_attributes(self):
        # Test writing face attributes to a ply, by reading
        # them back and asserting the written attributes array matches

        m = g.get_mesh('box.STL')
        test_1d_attribute = g.np.copy(m.face_angles[:, 0])
        test_nd_attribute = g.np.copy(m.face_angles)
        m.face_attributes['test_1d_attribute'] = test_1d_attribute
        m.face_attributes['test_nd_attribute'] = test_nd_attribute

        export = m.export(file_type='ply')
        reconstructed = g.wrapload(export, file_type='ply')

        face_attributes = reconstructed.metadata['_ply_raw']['face']['data']
        result_1d = face_attributes['test_1d_attribute']
        result_nd = face_attributes['test_nd_attribute']['f1']

        g.np.testing.assert_almost_equal(result_1d, test_1d_attribute)
        g.np.testing.assert_almost_equal(result_nd, test_nd_attribute)

        no_attr = m.export(file_type='ply', include_attributes=False)
        assert len(no_attr) < len(export)

    def test_cases(self):
        a = g.get_mesh('featuretype.STL')
        b = g.get_mesh('featuretype.ply')
        assert a.faces.shape == b.faces.shape

        # has mixed quads and triangles
        m = g.get_mesh('suzanne.ply')
        assert len(m.faces) > 0

    def test_ascii_color(self):
        mesh = g.trimesh.creation.box()
        en = g.wrapload(mesh.export(file_type='ply', encoding="ascii"),
                        file_type='ply')
        assert en.visual.kind is None

        color = [255, 0, 0, 255]
        mesh.visual.vertex_colors = color

        # try exporting and reloading raw
        eb = g.wrapload(mesh.export(file_type='ply'), file_type='ply')

        assert g.np.allclose(eb.visual.vertex_colors[0], color)
        assert eb.visual.kind == 'vertex'

        ea = g.wrapload(mesh.export(file_type='ply', encoding='ascii'),
                        file_type='ply')
        assert g.np.allclose(ea.visual.vertex_colors, color)
        assert ea.visual.kind == 'vertex'

    def test_empty_or_pointcloud(self):
        # demo files to check
        empty_files = ['ply_empty_ascii.ply',
                       'ply_empty_bin.ply',
                       'ply_empty_header.ply',
                       'ply_points_ascii.ply',
                       'ply_points_bin.ply']

        for empty_file in empty_files:
            e = g.get_mesh('emptyIO/' + empty_file)
            if 'empty' in empty_file:
                # result should be an empty scene
                try:
                    e.export(file_type='ply')
                except BaseException:
                    continue
                raise ValueError('should not export empty')
            elif 'points' in empty_file:
                # create export
                export = e.export(file_type='ply')
                reconstructed = g.wrapload(export, file_type='ply')

                # result should be a point cloud instance
                assert isinstance(e, g.trimesh.PointCloud)
                assert hasattr(e, 'vertices')
                # point cloud export should contain vertices
                assert isinstance(reconstructed, g.trimesh.PointCloud)
                assert hasattr(reconstructed, 'vertices')

    def test_blender_uv(self):
        # test texture coordinate loading for Blender exported ply files
        mesh_names = []

        # test texture coordinate loading for simple triangulated
        # Blender-export
        mesh_names.append('cube_blender_uv.ply')

        # same mesh but re-exported from meshlab as binary ply (and with
        # changed header)
        mesh_names.append('cube_blender_uv_meshlab.ply')

        # test texture coordinate loading for mesh with mixed quads and
        # triangles
        mesh_names.append('suzanne.ply')

        for mesh_name in mesh_names:
            m = g.get_mesh(mesh_name)
            assert hasattr(m, 'visual') and hasattr(m.visual, 'uv')
            assert m.visual.uv.shape[0] == m.vertices.shape[0]

    def test_uv_export(self):
        m = g.get_mesh("fuze.ply")
        assert hasattr(m, 'visual') and hasattr(m.visual, 'uv')
        assert m.visual.uv.shape[0] == m.vertices.shape[0]

        # create empty file to export to

        with g.TemporaryDirectory() as D:
            name = g.os.path.join(D, 'file.ply')

            # export should contain the uv data
            m.export(name)
            m2 = g.trimesh.load(name)

        assert hasattr(m2, 'visual') and hasattr(m2.visual, 'uv')
        assert g.np.allclose(m.visual.uv, m2.visual.uv)

    def test_fix_texture(self):
        # test loading of face indices when uv-coordinates are also contained
        m1 = g.get_mesh('plane.ply')
        m2 = g.get_mesh('plane_tri.ply')
        assert m1.faces.shape == (2, 3)
        assert m2.faces.shape == (2, 3)

    def test_texturefile(self):
        # try loading a PLY with texture
        m = g.get_mesh('fuze.ply')
        # run the checks to make sure fuze has the
        # correct number of vertices and has texture loaded
        g.check_fuze(m)

    def test_metadata(self):
        mesh = g.get_mesh('metadata.ply')

        assert (g.np.array([[12], [90]]) == mesh.metadata[
            '_ply_raw']['face']['data']['face_type']).all()


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class NormalsTest(g.unittest.TestCase):

    def test_vertex_normal(self):
        mesh = g.trimesh.creation.icosahedron()
        # the icosahedron is centered at zero, so the true vertex
        # normal is just a unit vector of the vertex position
        truth = g.trimesh.util.unitize(mesh.vertices)

        # force fallback to loop normal summing by passing None
        # as the sparse matrix
        normals = g.trimesh.geometry.mean_vertex_normals(
            len(mesh.vertices),
            mesh.faces,
            mesh.face_normals)
        assert g.np.allclose(normals - truth, 0.0)

        # make sure the automatic sparse matrix generation works
        normals = g.trimesh.geometry.mean_vertex_normals(
            len(mesh.vertices),
            mesh.faces,
            mesh.face_normals)
        assert g.np.allclose(normals - truth, 0.0)

        # make sure the Trimesh normals- related attributes
        # are wired correctly
        assert mesh.faces_sparse is not None
        assert mesh.vertex_normals.shape == mesh.vertices.shape
        assert g.np.allclose(mesh.vertex_normals - truth, 0.0)

    def test_weighted_vertex_normals(self):

        def compare_trimesh_to_groundtruth(
                mesh,
                truth, atol=g.trimesh.tol.merge):
            # force fallback to loop normal summing by passing None
            # as the sparse matrix
            normals = g.trimesh.geometry.weighted_vertex_normals(
                vertex_count=len(mesh.vertices),
                faces=mesh.faces,
                face_normals=mesh.face_normals,
                face_angles=mesh.face_angles)
            assert g.np.allclose(normals - truth, 0.0, atol=atol)

            # make sure the automatic sparse matrix generation works
            normals = g.trimesh.geometry.weighted_vertex_normals(
                len(mesh.vertices),
                mesh.faces,
                mesh.face_normals,
                mesh.face_angles)
            assert g.np.allclose(normals - truth, 0.0, atol=atol)

            # make sure the Trimesh normals- related attributes
            # are wired correctly
            assert mesh.faces_sparse is not None
            assert mesh.vertex_normals.shape == mesh.vertices.shape
            assert g.np.allclose(mesh.vertex_normals - truth, 0.0, atol=atol)

        # the icosahedron is centered at zero, so the true vertex
        # normal is just a unit vector of the vertex position
        ico_mesh = g.trimesh.creation.icosahedron()
        ico_truth = g.trimesh.util.unitize(ico_mesh.vertices)
        compare_trimesh_to_groundtruth(ico_mesh, ico_truth)

        # create a cube centered at zero, as with the icosahedron,
        # normals compute as unit vectors of the corner vertices
        # due to the triangulation of the box, this case would fail
        # with a simple face-normals-average as vertex-normal method
        box_mesh = g.trimesh.creation.box()
        box_truth = g.trimesh.util.unitize(box_mesh.vertices)
        compare_trimesh_to_groundtruth(box_mesh, box_truth)

        # load the fandisk model: a complex triangulated mesh with
        # smooth curved surfaces, corners and sharp creases
        # ground truth vertex normals were computed in MeshLab and
        # are included in the file
        fandisk_mesh = g.get_mesh('fandisk.obj')
        fandisk_truth = fandisk_mesh.vertex_normals
        # due to the limited precision in the MeshLab export,
        # we have to tweak the tolerance for the comparison a little
        compare_trimesh_to_groundtruth(fandisk_mesh, fandisk_truth, 0.0001)

        # see how we do with degenerate faces
        m = g.trimesh.creation.box()
        m.faces[0][0] = m.faces[0][1]
        norm = m.vertex_normals
        assert g.np.isfinite(norm).all()
        assert len(norm) == len(m.vertices)

        # vertices with every face intact
        mask = g.np.zeros(len(m.vertices), dtype=bool)
        mask[m.faces[0]] = False
        # it's a box so normals should all be unit vectors [1,1,1]
        assert g.np.allclose(g.np.abs(norm[mask]), (1.0 / 3.0) ** .5)

        # try with a deliberately broken sparse matrix to test looping path
        norm = g.trimesh.geometry.weighted_vertex_normals(
            vertex_count=len(m.vertices),
            faces=m.faces,
            face_normals=m.face_normals,
            face_angles=m.face_angles,
            use_loop=True)
        assert g.np.isfinite(norm).all()
        assert len(norm) == len(m.vertices)

        # every intact vertex should be away from box corner
        assert g.np.allclose(g.np.abs(norm[mask]), (1.0 / 3.0) ** .5)

    def test_face_normals(self):
        """
        Test automatic generation of face normals on mesh objects
        """
        mesh = g.trimesh.creation.icosahedron()
        assert mesh.face_normals.shape == mesh.faces.shape

        # normals should regenerate
        mesh.face_normals = None
        assert mesh.face_normals.shape == mesh.faces.shape

        # we shouldn't be able to assign stupid wrong values
        # even with nonzero and the right shape
        mesh.face_normals = g.np.ones_like(mesh.faces) * [0.0, 0.0, 1.0]
        assert not g.np.allclose(mesh.face_normals, [0.0, 0.0, 1.0])

        # setting normals to None should force recompute
        mesh.face_normals = None
        assert mesh.face_normals is not None
        assert not g.np.allclose(mesh.face_normals,
                                 [0.0, 0.0, 1.0])

        # setting face normals to zeros shouldn't work
        mesh.face_normals = g.np.zeros_like(mesh.faces)
        assert g.np.allclose(
            g.np.linalg.norm(mesh.face_normals, axis=1), 1.0)

    def test_merge(self):
        """
        Check merging with vertex normals
        """
        # no vertex merging
        m = g.get_mesh(
            'cube_compressed.obj',
            process=False)
        assert m.vertices.shape == (24, 3)
        assert m.faces.shape == (12, 3)
        assert g.np.isclose(m.volume, 8.0, atol=1e-4)

        # with normal-aware vertex merging
        m = g.get_mesh(
            'cube_compressed.obj',
            process=True)
        assert m.vertices.shape == (24, 3)
        assert m.faces.shape == (12, 3)
        assert g.np.isclose(m.volume, 8.0, atol=1e-4)

        # without considering normals should just be cube
        m.merge_vertices(merge_norm=True)
        assert m.vertices.shape == (8, 3)
        assert m.faces.shape == (12, 3)
        assert g.np.isclose(m.volume, 8.0, atol=1e-4)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class PosesTest(g.unittest.TestCase):

    def test_nonsampling_poses(self):
        mesh = g.trimesh.creation.icosahedron()

        # Compute the stable poses of the icosahedron
        trans, probs = mesh.compute_stable_poses()

        # Probabilities should all be 0.05 (20 faces)
        self.assertTrue(g.np.allclose(g.np.array(probs) - 0.05, 0.0))
        self.assertTrue(len(trans) == 20)
        self.assertTrue(len(probs) == 20)

    def test_multiple(self):
        for mesh in [g.trimesh.creation.icosahedron(),
                     g.get_mesh('unit_cube.STL')]:

            vectors = g.trimesh.util.grid_linspace([[0.0, 0], [1, 1.0]], 5)[1:]
            vectors = g.trimesh.unitize(g.np.column_stack(
                (vectors, g.np.ones(len(vectors)))))
            for vector, angle in zip(
                    vectors, g.np.linspace(0.0, g.np.pi, len(vectors))):
                matrix = g.trimesh.transformations.rotation_matrix(
                    angle, vector)

                copied = mesh.copy()
                copied.apply_transform(matrix)

                # Compute the stable poses of the icosahedron
                trans, probs = copied.compute_stable_poses()

                # we are only testing primitives with point symmetry
                # AKA 3 principal components of inertia are the same
                facet_count = len(mesh.facets)
                if facet_count == 0:
                    facet_count = len(mesh.faces)
                probability = 1.0 / float(facet_count)

                assert g.np.allclose(g.np.array(probs) - probability, 0.0)

    def test_round(self):
        mesh = g.trimesh.primitives.Cylinder(radius=1.0, height=10.0)

        transforms, probabilities = mesh.compute_stable_poses()
        transforms, probabilities = mesh.compute_stable_poses(n_samples=10)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class UnwrapTest(g.unittest.TestCase):

    def test_image(self):
        try:
            import xatlas  # noqa
        except BaseException:
            g.log.info('not testing unwrap as no `xatlas`')
            return
        a = g.get_mesh('bunny.ply', force="mesh")

        u = a.unwrap()
        assert u.visual.uv.shape == (len(u.vertices), 2)

        checkerboard = g.np.kron(
            [[1, 0] * 4, [0, 1] * 4] * 4, g.np.ones((10, 10)))
        try:
            from PIL import Image
        except BaseException:
            return

        image = Image.fromarray(
            (checkerboard * 255).astype(g.np.uint8))
        u = a.unwrap(image=image)
        # make sure image was attached correctly
        assert u.visual.material.image.size == image.size


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class BoundsTest(g.unittest.TestCase):

    def setUp(self):
        meshes = [g.get_mesh(i) for i in ['large_block.STL',
                                          'featuretype.STL']]
        self.meshes = g.np.append(meshes, list(g.get_meshes(5)))

    def test_obb_mesh(self):
        """
        Test the OBB functionality in attributes of Trimesh objects
        """
        for m in self.meshes:
            g.log.info('Testing OBB of %s', m.metadata['file_name'])
            for i in range(6):
                # on the first run through don't transform the points to see
                # if we succeed in the meshes original orientation
                matrix = g.np.eye(4)
                if i > 0:
                    # get repeatable transforms
                    matrix = g.transforms[i]
                    m.apply_transform(matrix)

                box_ext = m.bounding_box_oriented.primitive.extents.copy()
                box_t = m.bounding_box_oriented.primitive.transform.copy()

                # determinant of rotation should be 1.0
                assert g.np.isclose(g.np.linalg.det(box_t[:3, :3]), 1.0)

                m.apply_transform(g.np.linalg.inv(box_t))

                test = m.bounds / (box_ext / 2.0)
                test_ok = g.np.allclose(test, [[-1, -1, -1], [1, 1, 1]])
                if not test_ok:
                    g.log.error('bounds test failed %s',
                                str(test))
                assert test_ok

                m.apply_transform(matrix)
                m.apply_obb()

                # after applying the obb, the extents of the AABB
                # should be the same as the OBB
                close = g.np.allclose(m.bounding_box.extents,
                                      m.bounding_box_oriented.extents,
                                      rtol=1e-3,
                                      atol=1e-3)
                if not close:
                    # m.visual.face_colors = [200, 0, 0, 100]
                    # (m + m.bounding_box_oriented).show()
                    # from IPython import embed
                    # embed()
                    raise ValueError('OBB extents incorrect:\n{}\n{}'.format(
                        str(m.bounding_box.extents),
                        str(m.bounding_box_oriented.extents)))

            c = m.bounding_cylinder  # NOQA
            s = m.bounding_sphere  # NOQA
            p = m.bounding_primitive  # NOQA

    def test_obb_points(self):
        """
        Test OBB functions with raw points as input
        """
        for dimension in [3, 2]:
            for _i in range(25):
                points = g.random((10, dimension))
                to_origin, extents = g.trimesh.bounds.oriented_bounds(points)

                assert g.trimesh.util.is_shape(to_origin,
                                               (dimension + 1, dimension + 1))
                assert g.trimesh.util.is_shape(extents, (dimension,))

                transformed = g.trimesh.transform_points(points, to_origin)

                transformed_bounds = [transformed.min(axis=0),
                                      transformed.max(axis=0)]

                for j in transformed_bounds:
                    # assert that the points once our obb to_origin transform is applied
                    # has a bounding box centered on the origin
                    assert g.np.allclose(g.np.abs(j), extents / 2.0)

                extents_tf = g.np.diff(
                    transformed_bounds, axis=0).reshape(dimension)
                assert g.np.allclose(extents_tf, extents)

    def test_obb_coplanar_points(self):
        """
        Test OBB functions with coplanar points as input
        """
        for _i in range(5):
            points = g.np.zeros((5, 3))
            points[:, :2] = g.random((points.shape[0], 2))
            rot, _ = g.np.linalg.qr(g.random((3, 3)))
            points = g.np.matmul(points, rot)
            to_origin, extents = g.trimesh.bounds.oriented_bounds(points)

            assert g.trimesh.util.is_shape(to_origin, (4, 4))
            assert g.trimesh.util.is_shape(extents, (3,))

            transformed = g.trimesh.transform_points(points, to_origin)

            transformed_bounds = [transformed.min(axis=0),
                                  transformed.max(axis=0)]

            for j in transformed_bounds:
                # assert that the points once our obb to_origin transform is applied
                # has a bounding box centered on the origin
                assert g.np.allclose(g.np.abs(j), extents / 2.0)

            extents_tf = g.np.diff(transformed_bounds, axis=0).reshape(3)
            assert g.np.allclose(extents_tf, extents)

    def test_2D(self):
        for theta in g.np.linspace(0, g.np.pi * 2, 2000):
            # create some random rectangular-ish 2D points
            points = g.random((10, 2)) * [5, 1]

            # save the basic AABB of the points before rotation
            rectangle_pre = points.ptp(axis=0)

            # rotate them by an increment
            TR = g.trimesh.transformations.planar_matrix(theta=theta)
            points = g.trimesh.transform_points(points, TR)

            # find the OBB of the points
            T, rectangle = g.trimesh.bounds.oriented_bounds_2D(points)

            # apply the calculated OBB
            oriented = g.trimesh.transform_points(points, T)

            origin = oriented.min(axis=0) + oriented.ptp(axis=0) / 2.0

            # check to make sure the returned rectangle size is right
            assert g.np.allclose(oriented.ptp(axis=0), rectangle)
            # check to make sure the OBB consistently returns the
            # long axis in the same direction
            assert rectangle[0] > rectangle[1]
            # check to make sure result is actually returning an OBB
            assert g.np.allclose(origin, 0.0)
            # make sure OBB has less or same area as naive AABB
            assert g.np.product(rectangle) <= g.np.product(rectangle_pre)

    def test_cylinder(self):
        """
        Check bounding cylinders on basically a cuboid
        """
        # not rotationally symmetric
        mesh = g.get_mesh('featuretype.STL')

        height = 10.0
        radius = 1.0

        # spherical coordinates to loop through
        sphere = g.trimesh.util.grid_linspace(
            [[0, 0], [g.np.pi * 2, g.np.pi * 2]], 5)

        for s in sphere:
            T = g.trimesh.transformations.spherical_matrix(*s)
            p = g.trimesh.creation.cylinder(radius=radius,
                                            height=height,
                                            transform=T)
            assert g.np.isclose(radius,
                                p.bounding_cylinder.primitive.radius,
                                rtol=.01)
            assert g.np.isclose(height,
                                p.bounding_cylinder.primitive.height,
                                rtol=.01)

            # regular mesh should have the same bounding cylinder
            # regardless of transform
            copied = mesh.copy()
            copied.apply_transform(T)
            assert g.np.isclose(mesh.bounding_cylinder.volume,
                                copied.bounding_cylinder.volume,
                                rtol=.05)

    def test_random_cylinder(self):
        """
        Check exact cylinders with the bounding cylinder function.
        """
        for _i in range(20):
            # create a random cylinder
            c = g.trimesh.creation.cylinder(
                radius=1.0, height=10).permutate.transform()
            # bounding primitive should have same height and radius
            assert g.np.isclose(
                c.bounding_cylinder.primitive.height, 10, rtol=1e-6)
            assert g.np.isclose(
                c.bounding_cylinder.primitive.radius, 1, rtol=1e-6)
            # mesh is a cylinder, so center mass of bounding cylinder
            # should be exactly the same as the mesh center mass
            assert g.np.allclose(
                c.center_mass,
                c.bounding_cylinder.center_mass,
                rtol=1e-6)

    def test_bounding_egg(self):
        # create a distorted sphere mesh
        # center mass will be offset along Z
        i = g.trimesh.creation.icosphere()
        mask = i.vertices[:, 2] > 0.0
        i.vertices[:, 2][mask] *= 4.0

        # get a copy with a random transform
        p = i.permutate.transform()
        assert p.symmetry == 'radial'

        # find the bounding cylinder with this random transform
        r = p.bounding_cylinder

        # transformed height should match source mesh
        assert g.np.isclose(i.vertices[:, 2].ptp(),
                            r.primitive.height,
                            rtol=1e-6)
        # slightly inflated cylinder should contain all
        # vertices of the source mesh
        assert r.buffer(0.01).contains(p.vertices).all()

    def test_obb_order(self):
        # make sure our sorting and transform flipping of
        # OBB extents are working by checking against a box

        extents = [10, 2, 3.5]
        extents_ordered = g.np.sort(extents)

        for _i in range(100):
            # transform box randomly in rotation and translation
            mat = g.trimesh.transformations.random_rotation_matrix()
            # translate in box -100 : +100
            mat[:3, 3] = (g.random(3) - .5) * 200

            # source mesh to check
            b = g.trimesh.creation.box(extents=extents,
                                       transform=mat)

            # calculated OBB primitive
            obb = b.bounding_box_oriented

            # make sure extents returned were ordered
            assert g.np.allclose(obb.primitive.extents,
                                 extents_ordered)

            # make sure mesh isn't reversing windings
            assert g.np.isclose(obb.to_mesh().volume,
                                g.np.product(extents))

            # make sure OBB has the same bounds as the source mesh
            # since it is a box the AABB of the OBB should be
            # the same as the AABB of the source mesh (lol)
            assert g.np.allclose(obb.bounds,
                                 b.bounds)

            # unordered extents and transforms
            transform, extents = g.trimesh.bounds.oriented_bounds(
                b, ordered=False)
            assert g.np.allclose(g.np.sort(extents),
                                 extents_ordered)
            # create a box from the unordered OBB information
            box = g.trimesh.creation.box(
                extents=extents, transform=g.np.linalg.inv(transform))
            # make sure it is a real OBB too
            assert g.np.allclose(box.bounds, b.bounds)

    def test_bounds_tree(self):
        # test r-tree intersections
        for dimension in (2, 3):
            # create some (n, 2, 3) bounds
            bounds = g.np.array([[i.min(axis=0), i.max(axis=0)]
                                 for i in
                                 [g.random((4, dimension))
                                  for i in range(10)]])
            tree = g.trimesh.util.bounds_tree(bounds)
            for i, b in enumerate(bounds):
                assert i in set(tree.intersection(b.ravel()))
            # construct tree with per-row bounds
            tree = g.trimesh.util.bounds_tree(
                bounds.reshape((-1, dimension * 2)))
            for i, b in enumerate(bounds):
                assert i in set(tree.intersection(b.ravel()))

    def test_obb_corpus(self):
        # get some sample watertight meshes with nonzero volume
        min_volume = 0.1
        meshes = list(g.get_meshes(split=True,
                                   min_volume=min_volume,
                                   only_watertight=True))
        g.log.debug('loaded {} meshes'.format(len(meshes)))

        if g.PY3:
            # our models corpus should have 200+ models
            # some loaders aren't available for Python2
            assert len(meshes) > 200

        # triple check we have positive volume
        assert all(m.volume > min_volume for m in meshes)

        # compute the OBB for every mesh profiling
        with g.Profiler() as P:
            obb = [m.bounding_box_oriented for m in meshes]
        g.log.debug(P.output_text())

        # now loop through mesh-obb pairs and validate
        for m, o in zip(meshes, obb):
            # move the mesh into the OBB frame
            check = m.copy().apply_transform(
                g.np.linalg.inv(o.primitive.transform))
            # check the mesh bounds against the claimed OBB bounds
            half = o.primitive.extents / 2.0
            check_extents = g.np.array([-half, half])
            # check that the OBB does countain the mesh
            assert g.np.allclose(check.bounds, check_extents, rtol=1e-4)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
Check things related to STL files
"""
try:
    from . import generic as g
except BaseException:
    import generic as g


class STLTests(g.unittest.TestCase):

    def test_header(self):
        m = g.get_mesh('featuretype.STL')
        # make sure we have the right mesh
        assert g.np.isclose(m.volume, 11.627733431196749, atol=1e-6)

        # should have saved the header from the STL file
        assert len(m.metadata['header']) > 0

        # should have saved the STL face attributes
        assert len(m.face_attributes['stl']) == len(m.faces)
        assert len(m.faces) > 1000
        # add a non-correlated face attribute, which should be ignored
        m.face_attributes['nah'] = 10

        # remove all faces except three random ones
        m.update_faces([1, 3, 4])
        # faces and face attributes should be untouched
        assert len(m.faces) == 3
        assert len(m.face_attributes['stl']) == 3
        # attribute that wasn't len(m.faces) shouldn't have been touched
        assert m.face_attributes['nah'] == 10

    def test_attrib(self):
        m = g.get_mesh('featuretype.STL')

        len_vertices = len(m.vertices)
        # assign some random vertex attributes
        random = g.random(len(m.vertices))
        m.vertex_attributes['random'] = random
        m.vertex_attributes['nah'] = 20

        # should have saved the STL face attributes
        assert len(m.face_attributes['stl']) == len(m.faces)
        assert len(m.faces) > 1000
        # add a non-correlated face attribute, which should be ignored
        m.face_attributes['nah'] = 10

        # remove all faces except three random ones
        m.update_faces([1, 3, 4])
        # faces and face attributes should be untouched
        assert len(m.faces) == 3
        assert len(m.face_attributes['stl']) == 3
        # attribute that wasn't len(m.faces) shouldn't have been touched
        assert m.face_attributes['nah'] == 10

        # check all vertices are still in place
        assert m.vertex_attributes['nah'] == 20
        assert g.np.allclose(random, m.vertex_attributes['random'])
        assert len(m.vertices) == len_vertices

        # remove all vertices except four
        v_mask = [0, 1, 2, 3]
        m.update_vertices(v_mask)
        # make sure things are still correct
        assert m.vertex_attributes['nah'] == 20
        assert g.np.allclose(m.vertex_attributes['random'], random[v_mask])
        assert len(m.vertices) == len(v_mask)

    def test_ascii_multibody(self):
        s = g.get_mesh('multibody.stl')
        assert len(s.geometry) == 2
        assert set(s.geometry.keys()) == {'bodya', 'bodyb'}

    def test_empty(self):
        # demo files to check
        empty_files = ['stl_empty_ascii.stl',
                       'stl_empty_bin.stl']

        for empty_file in empty_files:
            e = g.get_mesh('emptyIO/' + empty_file)

            # result should be an empty scene without vertices
            assert isinstance(e, g.trimesh.Scene)
            assert not hasattr(e, 'vertices')

            # create export
            try:
                e.export(file_type='ply')
            except BaseException:
                return
            raise ValueError("Shouldn't export empty scenes!")

    def test_vertex_order(self):
        for stl in ['featuretype.STL',
                    'ADIS16480.STL',
                    '1002_tray_bottom.STL']:
            # removing doubles should respect the vertex order
            m_raw = g.get_mesh(stl, process=False)
            m_proc = g.get_mesh(stl, process=True,
                                keep_vertex_order=True)

            verts_raw = g.trimesh.grouping.hashable_rows(m_raw.vertices)
            verts_proc = g.trimesh.grouping.hashable_rows(m_proc.vertices)

            # go through all processed verts
            # find index in unprocessed mesh
            idxs = []
            for vert in verts_proc:
                idxs.append(g.np.where(verts_raw == vert)[0][0])

            # indices should be increasing
            assert (g.np.diff(idxs) > 0).all()

            tris_raw = m_raw.triangles
            tris_proc = m_proc.triangles

            # of course mesh needs to have same faces as before
            assert g.np.allclose(
                g.np.sort(tris_raw, axis=0),
                g.np.sort(tris_proc, axis=0))

    def test_ascii_keyword(self):
        m = g.get_mesh('ascii.stl.zip', force='mesh')
        assert m.is_watertight


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import triangle  # NOQA
    has_triangle = True
except ImportError:
    g.log.warning('Not testing extrude primitives!')
    has_triangle = False


class PrimitiveTest(g.unittest.TestCase):

    def setUp(self):
        self.primitives = []

        # do it with a flag in case there is more than one ImportError
        if has_triangle:
            e = g.trimesh.primitives.Extrusion()
            e.primitive.polygon = g.trimesh.path.polygons.random_polygon()
            e.primitive.height = 1.0
            self.primitives.append(e)

            self.primitives.append(
                g.trimesh.primitives.Extrusion(
                    polygon=g.trimesh.path.polygons.random_polygon(),
                    height=293292.322))

            self.primitives.append(e.buffer(.25))

            self.primitives.append(
                g.trimesh.primitives.Extrusion(
                    polygon=g.Point([0, 0]).buffer(.5),
                    height=1.0))

            self.primitives.append(
                g.trimesh.primitives.Extrusion(
                    polygon=g.Point([0, 0]).buffer(.5),
                    height=-1.0))

        self.primitives.append(g.trimesh.primitives.Sphere())
        self.primitives.append(g.trimesh.primitives.Sphere(center=[0, 0, 100],
                                                           radius=10.0,
                                                           subdivisions=5))
        self.primitives.append(g.trimesh.primitives.Box())
        self.primitives.append(
            g.trimesh.primitives.Box(
                center=[102.20, 0, 102.0],
                extents=[29, 100, 1000]))
        self.primitives.append(g.trimesh.primitives.Box(
            extents=[10, 20, 30],
            transform=g.trimesh.transformations.random_rotation_matrix()))

        self.primitives.append(g.trimesh.primitives.Cylinder())
        self.primitives.append(g.trimesh.primitives.Cylinder(radius=10,
                                                             height=1,
                                                             sections=40))

        self.primitives.append(g.trimesh.primitives.Capsule())
        self.primitives.append(g.trimesh.primitives.Capsule(radius=1.5,
                                                            height=10))

    def test_scaling(self):
        # try a simple scaling test
        p = g.trimesh.primitives.Sphere(radius=1.0)
        m = p.to_mesh()
        assert g.np.allclose(p.extents, 2.0)
        assert g.np.allclose(m.extents, 2.0)
        p.apply_scale(0.5)
        m.apply_scale(0.5)
        assert g.np.isclose(p.primitive.radius, 0.5)
        assert g.np.allclose(p.extents, 1.0)
        assert g.np.allclose(m.extents, 1.0)
        assert g.np.allclose(p.extents, m.extents, atol=1e-3)

        p = g.trimesh.primitives.Box()
        p.apply_translation([0.5, 0.5, 0.5])
        p.apply_scale(5.0)
        assert g.np.allclose(p.bounds, [[0, 0, 0], [5, 5, 5]])

        try:
            raised = False
            p.apply_scale([2, 3, 4])
        except BaseException:
            raised = True
        if not raised:
            raise ValueError('primitives should raise on non-uniform scaling')

        # now try with more complicated generated data
        prims = [g.trimesh.primitives.Sphere(radius=1.0),
                 g.trimesh.primitives.Sphere(radius=112.007),
                 g.trimesh.primitives.Cylinder(radius=1.0,
                                               height=10.0),
                 g.trimesh.primitives.Box(),
                 g.trimesh.primitives.Box(extents=[12, 32, 31]),
                 g.trimesh.primitives.Cylinder(radius=1.1212,
                                               height=0.001),
                 g.trimesh.primitives.Capsule(radius=1.0,
                                              height=7.0)]

        for original in prims:
            perm = [original,
                    original.copy(),
                    original.copy(),
                    original.copy()]
            # try with a simple translation
            perm[1].primitive.transform = g.tf.translation_matrix([0, 0, 7])
            perm[2].apply_transform(
                g.tf.rotation_matrix(g.np.pi / 4, [0, 0, 1]))
            # try with a gnarly rotation
            perm[3].primitive.transform = g.tf.random_rotation_matrix(
                translate=1000)

            fields = set(dir(original.primitive))
            ori_radius, ori_height = None, None
            if 'radius' in fields:
                ori_radius = original.primitive.radius
            if 'height' in fields:
                ori_height = original.primitive.height

            for scale in [1e-2, 0.123, 0.5, 100.2]:
                for po in perm:
                    # converting to mesh will do all scaling
                    # and transformations on simple discrete
                    # copy of primitive and should match with
                    # only tesselation differences
                    p = po.copy()
                    m = p.to_mesh()

                    # make sure we have the types we think we do
                    assert isinstance(p, g.trimesh.primitives._Primitive)
                    assert isinstance(m, g.trimesh.Trimesh)

                    assert g.np.allclose(p.extents, m.extents)

                    p.apply_scale(scale)
                    m.apply_scale(scale)

                    # matrix should never have scale
                    assert g.tf.is_rigid(p.primitive.transform)

                    if ori_radius is not None:
                        assert g.np.isclose(p.primitive.radius,
                                            ori_radius * scale)
                    if ori_height is not None:
                        assert g.np.isclose(p.primitive.height,
                                            ori_height * scale)

                    # should be the same size
                    assert g.np.allclose(
                        p.extents, m.extents, atol=1e-3 * scale)
                    # should be in the same place
                    assert g.np.allclose(p.bounds, m.bounds, atol=1e-3 * scale)

    def test_mesh_schema(self):
        # this schema should define every primitive.
        schema = g.trimesh.resources.get_schema(
            'primitive/trimesh.schema.json')
        # make sure a mesh passes the schema
        m = g.trimesh.creation.box()
        g.jsonschema.validate(m.to_dict(), schema)

    def test_primitives(self):

        kind = set([i.__class__.__name__
                    for i in self.primitives])
        # make sure our test data has every primitive
        kinds = {'Box', 'Capsule', 'Cylinder', 'Sphere'}
        if has_triangle:
            kinds.add('Extrusion')
        assert kind == kinds

        # this schema should define every primitive.
        schema = g.trimesh.resources.get_schema(
            'primitive/primitive.schema.json')

        for primitive in self.primitives:

            # convert to a dict
            d = primitive.to_dict()
            # validate the output of the to-dict method
            g.jsonschema.validate(d, schema)

            # just triple-check that we have a transform
            # this should have been validated by the schema
            assert g.np.shape(d['transform']) == (4, 4)
            assert g.trimesh.transformations.is_rigid(
                d['transform'])
            # make sure the value actually json-dumps
            assert len(g.json.dumps(d)) > 0

            # make sure faces and vertices are correct
            assert g.trimesh.util.is_shape(primitive.faces,
                                           (-1, 3))
            assert g.trimesh.util.is_shape(primitive.vertices,
                                           (-1, 3))
            # check dtype of faces and vertices
            assert primitive.faces.dtype.kind == 'i'
            assert primitive.vertices.dtype.kind == 'f'

            assert primitive.volume > 0.0
            assert primitive.area > 0.0

            # convert to base class trimesh
            as_mesh = primitive.to_mesh()

            assert as_mesh.volume > 0.0
            assert as_mesh.area > 0.0

            assert g.np.allclose(primitive.extents,
                                 as_mesh.extents)
            assert g.np.allclose(primitive.bounds,
                                 as_mesh.bounds)

            assert g.np.isclose(primitive.volume,
                                as_mesh.volume,
                                rtol=.05)
            assert g.np.isclose(primitive.area,
                                as_mesh.area,
                                rtol=.05)

            assert primitive.is_winding_consistent
            assert primitive.is_watertight
            assert as_mesh.is_winding_consistent
            assert as_mesh.is_watertight

            # check that overload of dir worked
            assert len([i for i in
                        dir(primitive.primitive)
                        if '_' not in i]) > 0
            if hasattr(primitive, 'direction'):
                assert primitive.direction.shape == (3,)

            centroid = primitive.centroid.copy()
            translation = [0, 0, 5]
            primitive.apply_translation(translation)

            # centroid should have translated correctly
            assert g.np.allclose(primitive.centroid - centroid,
                                 translation)

    def test_sample(self):
        transform = g.trimesh.transformations.random_rotation_matrix()
        box = g.trimesh.primitives.Box(transform=transform,
                                       extents=[20, 10, 100])
        for kwargs in [{'step': 8},
                       {'step': [10, .4, 10]},
                       {'count': 8},
                       {'count': [10, 3, 5]}]:
            grid = box.sample_grid(**kwargs)
            assert g.trimesh.util.is_shape(grid, (-1, 3))
            assert (box.nearest.signed_distance(grid) > -1e-6).all()

    def test_box(self):
        """
        Test the setter on box primitives
        """
        start = [20, 10, 100]
        box = g.trimesh.primitives.Box(extents=start)
        assert g.np.allclose(box.primitive.extents,
                             start)
        assert g.np.allclose(box.extents,
                             start)
        if g.has_path:
            # check to see if outline function works
            assert g.np.allclose(box.as_outline().extents, start)

    def test_cyl_buffer(self):
        # test our inflation of cylinder primitives
        c = g.trimesh.primitives.Cylinder(
            radius=1.0,
            height=10.0,
            transform=g.trimesh.transformations.random_rotation_matrix())
        # inflate cylinder
        b = c.buffer(1.0)
        assert g.np.isclose(b.primitive.height, 12.0)
        assert g.np.isclose(b.primitive.radius, 2.0)
        # should contain all vertices of source mesh
        assert b.contains(c.vertices).all()
        # should contain line segment
        assert b.contains(c.segment).all()

    def test_transform_attribute(self):
        for primitive in self.primitives:
            assert hasattr(primitive, 'transform')

            assert g.trimesh.util.is_shape(primitive.transform,
                                           (4, 4))

            if hasattr(primitive.primitive, 'center'):
                assert g.np.allclose(primitive.primitive.center,
                                     primitive.transform[:3, 3])

    def test_sphere_center(self):
        s = g.trimesh.primitives.Sphere(
            center=[0, 0, 100], radius=10.0, subdivisions=5)
        assert g.np.allclose(s.center, [0, 0, 100])

        s.center = [1, 1, 1]
        assert g.np.allclose(s.center, [1, 1, 1])

        s.center[:2] = [0, 3]
        assert g.np.allclose(s.center, [0, 3, 1])


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()
